import renderdoc as rd
import rdtest


class GL_Buffer_Resizing(rdtest.TestCase):
    demos_test_name = 'GL_Buffer_Resizing'
    demos_frame_cap = 10

    def check_capture(self):
        postvs_ref = {
            0: {
                'vtx': 0,
                'idx': 0,
                'gl_Position': [-0.5, -0.5, 0.0, 1.0],
                'v2f_block.pos': [-0.5, -0.5, 0.0, 1.0],
                'v2f_block.col': [0.0, 1.0, 0.0, 1.0],
                'v2f_block.uv': [0.0, 0.0, 0.0, 1.0],
            },
            1: {
                'vtx': 1,
                'idx': 1,
                'gl_Position': [0.0, 0.5, 0.0, 1.0],
                'v2f_block.pos': [0.0, 0.5, 0.0, 1.0],
                'v2f_block.col': [0.0, 1.0, 0.0, 1.0],
                'v2f_block.uv': [0.0, 1.0, 0.0, 1.0],
            },
            2: {
                'vtx': 2,
                'idx': 2,
                'gl_Position': [0.5, -0.5, 0.0, 1.0],
                'v2f_block.pos': [0.5, -0.5, 0.0, 1.0],
                'v2f_block.col': [0.0, 1.0, 0.0, 1.0],
                'v2f_block.uv': [1.0, 0.0, 0.0, 1.0],
            },
        }

        draw = self.get_first_draw()

        idx = 0

        while True:
            draw: rd.DrawcallDescription = self.find_draw('glDraw', draw.eventId+1)

            if draw is None:
                break

            self.controller.SetFrameEvent(draw.eventId, True)

            self.check_triangle(out=draw.outputs[0])

            postvs_data = self.get_postvs(draw, rd.MeshDataStage.VSOut, 0, draw.numIndices)

            self.check_mesh_data(postvs_ref, postvs_data)

            idx = idx + 1
            rdtest.log.success('Draw {} at {} is correct'.format(idx, draw.eventId))



import rdtest
import renderdoc as rd


class GL_CBuffer_Zoo(rdtest.TestCase):
    demos_test_name = 'GL_CBuffer_Zoo'

    def check_capture(self):
        draw = self.find_draw("Draw")

        self.check(draw is not None)

        self.controller.SetFrameEvent(draw.eventId, False)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        stage = rd.ShaderStage.Pixel
        cbuf: rd.BoundCBuffer = pipe.GetConstantBuffer(stage, 0, 0)

        var_check = rdtest.ConstantBufferChecker(
            self.controller.GetCBufferVariableContents(pipe.GetGraphicsPipelineObject(),
                                                       pipe.GetShader(stage),
                                                       pipe.GetShaderEntryPoint(stage), 0,
                                                       cbuf.resourceId, cbuf.byteOffset, cbuf.byteSize))

        # For more detailed reference for the below checks, see the commented definition of the cbuffer
        # in the shader source code in the demo itself

        # vec4 a;
        var_check.check('a').cols(4).rows(1).value([0.0, 1.0, 2.0, 3.0])

        # vec3 b;
        var_check.check('b').cols(3).rows(1).value([4.0, 5.0, 6.0])

        # vec2 c; vec2 d;
        var_check.check('c').cols(2).rows(1).value([8.0, 9.0])
        var_check.check('d').cols(2).rows(1).value([10.0, 11.0])

        # float e; vec3 f;
        var_check.check('e').cols(1).rows(1).value([12.0])
        var_check.check('f').cols(3).rows(1).value([16.0, 17.0, 18.0])

        # vec4 dummy0;
        var_check.check('dummy0')

        # float j; vec2 k;
        var_check.check('j').cols(1).rows(1).value([24.0])
        var_check.check('k').cols(2).rows(1).value([26.0, 27.0])

        # vec2 l; float m;
        var_check.check('l').cols(2).rows(1).value([28.0, 29.0])
        var_check.check('m').cols(1).rows(1).value([30.0])

        # float n[4];
        var_check.check('n').cols(0).rows(0).arraySize(4).members({
            0: lambda x: x.cols(1).rows(1).value([32.0]),
            1: lambda x: x.cols(1).rows(1).value([36.0]),
            2: lambda x: x.cols(1).rows(1).value([40.0]),
            3: lambda x: x.cols(1).rows(1).value([44.0]),
        })

        # vec4 dummy1;
        var_check.check('dummy1')

        # float o[4];
        var_check.check('o').cols(0).rows(0).arraySize(4).members({
            0: lambda x: x.cols(1).rows(1).value([52.0]),
            1: lambda x: x.cols(1).rows(1).value([56.0]),
            2: lambda x: x.cols(1).rows(1).value([60.0]),
            3: lambda x: x.cols(1).rows(1).value([64.0]),
        })

        # float p;
        var_check.check('p').cols(1).rows(1).value([68.0])

        # vec4 dummy2;
        var_check.check('dummy2')

        # column_major vec4x4 q;
        var_check.check('q').cols(4).rows(4).column_major().value([76.0, 80.0, 84.0, 88.0,
                                                                   77.0, 81.0, 85.0, 89.0,
                                                                   78.0, 82.0, 86.0, 90.0,
                                                                   79.0, 83.0, 87.0, 91.0])

        # row_major vec4x4 r;
        var_check.check('r').cols(4).rows(4).row_major().value([92.0, 93.0, 94.0, 95.0,
                                                                96.0, 97.0, 98.0, 99.0,
                                                                100.0, 101.0, 102.0, 103.0])

        # column_major vec4x3 s;
        var_check.check('s').cols(4).rows(3).column_major().value([108.0, 112.0, 116.0, 120.0,
                                                                   109.0, 113.0, 117.0, 121.0,
                                                                   110.0, 114.0, 118.0, 122.0])

        # vec4 dummy3;
        var_check.check('dummy3')

        # row_major vec4x3 t;
        var_check.check('t').cols(4).rows(3).row_major().value([128.0, 129.0, 130.0, 131.0,
                                                                132.0, 133.0, 134.0, 135.0,
                                                                136.0, 137.0, 138.0, 139.0])

        # vec4 dummy4;
        var_check.check('dummy4')

        # column_major vec2x3 u;
        var_check.check('u').cols(3).rows(2).column_major().value([144.0, 148.0, 152.0,
                                                                   145.0, 149.0, 153.0])

        # vec4 dummy5;
        var_check.check('dummy5')

        # row_major vec3x2 v;
        var_check.check('v').cols(3).rows(2).row_major().value([160.0, 161.0, 162.0,
                                                                164.0, 165.0, 166.0])

        # vec4 dummy6;
        var_check.check('dummy6')

        # column_major vec3x2 w;
        var_check.check('w').cols(2).rows(2).column_major().value([172.0, 176.0,
                                                                   173.0, 177.0])

        # vec4 dummy7;
        var_check.check('dummy7')

        # row_major vec3x2 x;
        var_check.check('x').cols(2).rows(2).row_major().value([184.0, 185.0,
                                                                188.0, 189.0])

        # vec4 dummy8;
        var_check.check('dummy8')

        # row_major vec2x2 y;
        var_check.check('y').cols(2).rows(2).row_major().value([196.0, 197.0,
                                                                200.0, 201.0])

        # float z;
        var_check.check('z').cols(1).rows(1).value([204.0])

        # vec4 dummy9;
        var_check.check('dummy9')

        # vec4 multiarray[3][2];
        var_check.check('multiarray').cols(0).rows(0).arraySize(3).members({
            0: lambda x: x.cols(0).rows(0).arraySize(2).members({
                0: lambda y: y.cols(4).rows(1).value([228.0, 229.0, 230.0, 231.0]),
                1: lambda y: y.cols(4).rows(1).value([232.0, 233.0, 234.0, 235.0]),
            }),
            1: lambda x: x.cols(0).rows(0).arraySize(2).members({
                0: lambda y: y.cols(4).rows(1).value([236.0, 237.0, 238.0, 239.0]),
                1: lambda y: y.cols(4).rows(1).value([240.0, 241.0, 242.0, 243.0]),
            }),
            2: lambda x: x.cols(0).rows(0).arraySize(2).members({
                0: lambda y: y.cols(4).rows(1).value([244.0, 245.0, 246.0, 247.0]),
                1: lambda y: y.cols(4).rows(1).value([248.0, 249.0, 250.0, 251.0]),
            }),
        })

        # struct vec3_1 { vec3 a; float b; };
        # struct nested { vec3_1 a; vec4 b[4]; vec3_1 c[4]; };
        # nested structa[2];
        var_check.check('structa').cols(0).rows(0).arraySize(2).members({
            # structa[0]
            0: lambda s: s.cols(0).rows(0).structSize(3).members({
                'a': lambda x: x.cols(0).rows(0).structSize(2).members({
                    'a': lambda y: y.cols(3).rows(1).value([252.0, 253.0, 254.0]),
                    'b': lambda y: y.cols(1).rows(1).value([255.0]),
                }),
                'b': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(4).rows(1).value([256.0, 257.0, 258.0, 259.0]),
                    1: lambda y: y.cols(4).rows(1).value([260.0, 261.0, 262.0, 263.0]),
                    2: lambda y: y.cols(4).rows(1).value([264.0, 265.0, 266.0, 267.0]),
                    3: lambda y: y.cols(4).rows(1).value([268.0, 269.0, 270.0, 271.0]),
                }),
                'c': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([272.0, 273.0, 274.0]),
                        'b': lambda z: z.cols(1).rows(1).value([275.0]),
                    }),
                    1: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([276.0, 277.0, 278.0]),
                        'b': lambda z: z.cols(1).rows(1).value([279.0]),
                    }),
                    2: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([280.0, 281.0, 282.0]),
                        'b': lambda z: z.cols(1).rows(1).value([283.0]),
                    }),
                    3: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([284.0, 285.0, 286.0]),
                        'b': lambda z: z.cols(1).rows(1).value([287.0]),
                    }),
                }),
            }),
            # structa[1]
            1: lambda s: s.cols(0).rows(0).structSize(3).members({
                'a': lambda x: x.cols(0).rows(0).structSize(2).members({
                    'a': lambda y: y.cols(3).rows(1).value([288.0, 289.0, 290.0]),
                    'b': lambda y: y.cols(1).rows(1).value([291.0]),
                }),
                'b': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(4).rows(1).value([292.0, 293.0, 294.0, 295.0]),
                    1: lambda y: y.cols(4).rows(1).value([296.0, 297.0, 298.0, 299.0]),
                    2: lambda y: y.cols(4).rows(1).value([300.0, 301.0, 302.0, 303.0]),
                    3: lambda y: y.cols(4).rows(1).value([304.0, 305.0, 306.0, 307.0]),
                }),
                'c': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([308.0, 309.0, 310.0]),
                        'b': lambda z: z.cols(1).rows(1).value([311.0]),
                    }),
                    1: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([312.0, 313.0, 314.0]),
                        'b': lambda z: z.cols(1).rows(1).value([315.0]),
                    }),
                    2: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([316.0, 317.0, 318.0]),
                        'b': lambda z: z.cols(1).rows(1).value([319.0]),
                    }),
                    3: lambda y: y.cols(0).rows(0).structSize(2).members({
                        'a': lambda z: z.cols(3).rows(1).value([320.0, 321.0, 322.0]),
                        'b': lambda z: z.cols(1).rows(1).value([323.0]),
                    }),
                }),
            }),
        })

        # column_major mat2x3 ac;
        var_check.check('ac').cols(2).rows(3).column_major().value([324.0, 328.0,
                                                                    325.0, 329.0,
                                                                    326.0, 330.0])

        # row_major mat2x3 ad;
        var_check.check('ad').cols(2).rows(3).row_major().value([332.0, 333.0,
                                                                 336.0, 337.0,
                                                                 340.0, 341.0])

        # column_major mat2x3 ae[2];
        var_check.check('ae').cols(0).rows(0).arraySize(2).members({
            0: lambda x: x.cols(2).rows(3).column_major().value([344.0, 348.0,
                                                                 345.0, 349.0,
                                                                 346.0, 350.0]),
            1: lambda x: x.cols(2).rows(3).column_major().value([352.0, 356.0,
                                                                 353.0, 357.0,
                                                                 354.0, 358.0]),
        })

        # row_major mat2x3 af[2];
        var_check.check('af').cols(0).rows(0).arraySize(2).members({
            0: lambda x: x.cols(2).rows(3).row_major().value([360.0, 361.0,
                                                              364.0, 365.0,
                                                              368.0, 369.0]),
            1: lambda x: x.cols(2).rows(3).row_major().value([372.0, 373.0,
                                                              376.0, 377.0,
                                                              380.0, 381.0]),
        })

        # vec2 dummy10;
        var_check.check('dummy10')

        # row_major mat2x2 ag;
        var_check.check('ag').cols(2).rows(2).row_major().value([388.0, 389.0,
                                                                 392.0, 393.0])

        # vec2 dummy12;
        var_check.check('dummy11')

        # column_major float2x2 ah;
        var_check.check('ah').cols(2).rows(2).column_major().value([400.0, 404.0,
                                                                    401.0, 405.0])

        # row_major mat2x2 ai[2];
        var_check.check('ai').rows(0).cols(0).arraySize(2).members({
            0: lambda x: x.cols(2).rows(2).row_major().value([408.0, 409.0,
                                                              412.0, 413.0]),
            1: lambda x: x.cols(2).rows(2).row_major().value([416.0, 417.0,
                                                              420.0, 421.0]),
        })

        # column_major mat2x2 aj[2];
        var_check.check('aj').rows(0).cols(0).arraySize(2).members({
            0: lambda x: x.cols(2).rows(2).column_major().value([424.0, 428.0,
                                                                 425.0, 429.0]),
            1: lambda x: x.cols(2).rows(2).column_major().value([432.0, 436.0,
                                                                 433.0, 437.0]),
        })

        # struct nested_with_padding
        # {
        #   float a; // vec3 padding
        #   vec4 b;
        #   float c; // vec3 padding
        #   vec3 d[4]; // float padding after each one
        # };
        # nested_with_padding ak[2];
        var_check.check('ak').rows(0).cols(0).arraySize(2).members({
            # ak[0]
            0: lambda s: s.rows(0).cols(0).structSize(4).members({
                'a': lambda y: y.rows(1).cols(1).value([440.0]),
                'b': lambda y: y.rows(1).cols(4).value([444.0, 445.0, 446.0, 447.0]),
                'c': lambda y: y.rows(1).cols(1).value([448.0]),
                'd': lambda x: x.rows(0).cols(0).arraySize(4).members({
                    0: lambda z: z.rows(1).cols(3).value([452.0, 453.0, 454.0]),
                    1: lambda z: z.rows(1).cols(3).value([456.0, 457.0, 458.0]),
                    2: lambda z: z.rows(1).cols(3).value([460.0, 461.0, 462.0]),
                    3: lambda z: z.rows(1).cols(3).value([464.0, 465.0, 466.0]),
                }),
            }),
            # ak[1]
            1: lambda s: s.rows(0).cols(0).structSize(4).members({
                'a': lambda y: y.rows(1).cols(1).value([468.0]),
                'b': lambda y: y.rows(1).cols(4).value([472.0, 473.0, 474.0, 475.0]),
                'c': lambda y: y.rows(1).cols(1).value([476.0]),
                'd': lambda x: x.rows(0).cols(0).arraySize(4).members({
                    0: lambda z: z.rows(1).cols(3).value([480.0, 481.0, 482.0]),
                    1: lambda z: z.rows(1).cols(3).value([484.0, 485.0, 486.0]),
                    2: lambda z: z.rows(1).cols(3).value([488.0, 489.0, 490.0]),
                    3: lambda z: z.rows(1).cols(3).value([492.0, 493.0, 494.0]),
                }),
            }),
        })

        # float2 dummy12;
        var_check.check('dummy12')

        # float al;
        var_check.check('al').rows(1).cols(1).value([500.0])

        # struct float2_struct
        # {
        #   float x, y;
        # };
        # float2_struct am;
        var_check.check('am').rows(0).cols(0).members({
            'x': lambda y: y.rows(1).cols(1).value([504.0]),
            'y': lambda y: y.rows(1).cols(1).value([505.0]),
        })

        # float an;
        var_check.check('an').rows(1).cols(1).value([508.0])

        # float4 dummy13[2];
        var_check.check('dummy13')

        # misaligned_struct ao[2];
        var_check.check('ao').rows(0).cols(0).arraySize(2).members({
            # ao[0]
            0: lambda s: s.rows(0).cols(0).structSize(2).members({
                'a': lambda y: y.rows(1).cols(4).value([520.0, 521.0, 522.0, 523.0]),
                'b': lambda y: y.rows(1).cols(2).value([524.0, 525.0]),
            }),
            1: lambda s: s.rows(0).cols(0).structSize(2).members({
                'a': lambda y: y.rows(1).cols(4).value([528.0, 529.0, 530.0, 531.0]),
                'b': lambda y: y.rows(1).cols(2).value([532.0, 533.0]),
            }),
        })

        # float4 test;
        var_check.check('test').rows(1).cols(4).value([536.0, 537.0, 538.0, 539.0])

        # to save duplicating if this array changes, we calculate out from the start, as the array is tightly packed
        base = 540.0

        exp_vals = lambda wi,yi,xi: [base + wi * 24.0 + yi * 8.0 + xi * 4.0 + c * 1.0 for c in range(0,4)]

        # vec4 multiarray2[4][3][2];
        var_check.check('multiarray2').cols(0).rows(0).arraySize(4).members({
            0: lambda w: w.cols(0).rows(0).arraySize(3).members({
                0: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(0, 0, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(0, 0, 1)),
                }),
                1: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(0, 1, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(0, 1, 1)),
                }),
                2: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(0, 2, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(0, 2, 1)),
                }),
            }),
            1: lambda w: w.cols(0).rows(0).arraySize(3).members({
                0: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(1, 0, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(1, 0, 1)),
                }),
                1: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(1, 1, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(1, 1, 1)),
                }),
                2: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(1, 2, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(1, 2, 1)),
                }),
            }),
            2: lambda w: w.cols(0).rows(0).arraySize(3).members({
                0: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(2, 0, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(2, 0, 1)),
                }),
                1: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(2, 1, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(2, 1, 1)),
                }),
                2: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(2, 2, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(2, 2, 1)),
                }),
            }),
            3: lambda w: w.cols(0).rows(0).arraySize(3).members({
                0: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(3, 0, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(3, 0, 1)),
                }),
                1: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(3, 1, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(3, 1, 1)),
                }),
                2: lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda y: y.cols(4).rows(1).value(exp_vals(3, 2, 0)),
                    1: lambda y: y.cols(4).rows(1).value(exp_vals(3, 2, 1)),
                }),
            }),
        })

        var_check.done()

        rdtest.log.success("CBuffer variables are as expected")

        self.check_pixel_value(pipe.GetOutputTargets()[0].resourceId, 0.5, 0.5, [536.1, 537.0, 538.0, 539.0])

        rdtest.log.success("Picked value is as expected")

        cbuf: rd.BoundCBuffer = pipe.GetConstantBuffer(stage, 1, 0)

        var_check = rdtest.ConstantBufferChecker(
            self.controller.GetCBufferVariableContents(pipe.GetGraphicsPipelineObject(),
                                                       pipe.GetShader(stage),
                                                       pipe.GetShaderEntryPoint(stage), 1,
                                                       cbuf.resourceId, cbuf.byteOffset, cbuf.byteSize))

        # For bare uniforms we have partial data - only values used in the shader need to get assigned locations and
        # some drivers are aggressive about stripping any others. Only uniforms with locations get upload values.
        # Hence some of these checks don't check for the value - that means they might not be present

        # vec4 A;
        var_check.check('A').cols(4).rows(1).value([10.0, 20.0, 30.0, 40.0])

        # vec2 B;
        var_check.check('B').cols(2).rows(1).value([50.0, 60.0])

        # vec3 C;
        var_check.check('C').cols(3).rows(1).value([70.0, 80.0, 90.0])

        # mat2x3 D;
        var_check.check('D').cols(2).rows(3).value([100.0, 130.0,
                                                    110.0, 140.0,
                                                    120.0, 150.0])

        # float E[3];
        var_check.check('E').cols(0).rows(0).arraySize(3).members({
            0: lambda x: x.cols(1).rows(1).value([160.0]),
            1: lambda x: x.cols(1).rows(1).value([170.0]),
            2: lambda x: x.cols(1).rows(1).value([180.0]),
        })

        # vec4 F[3][2][2];
        # Multidimensional arrays are represented as structs with N members
        # Due to lacking reflection, we skip structSize() checks, but check everything else if it's present (if it's
        # not, it will be skipped)
        var_check.check('F').cols(0).rows(0).members({
            '[0]': lambda x: x.cols(0).rows(0).members({
                '[0]': lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda x: x.cols(4).rows(1).value([190.0, 200.0, 210.0, 220.0]),
                    1: lambda x: x.cols(4).rows(1).value([230.0, 240.0, 250.0, 260.0]),
                }),
                '[1]': lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda x: x.cols(4).rows(1).value([270.0, 280.0, 290.0, 300.0]),
                    1: lambda x: x.cols(4).rows(1).value([310.0, 320.0, 330.0, 340.0]),
                }),
            }),
            '[1]': lambda x: x.cols(0).rows(0).members({
                '[0]': lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda x: x.cols(4).rows(1).value([350.0, 360.0, 370.0, 380.0]),
                    1: lambda x: x.cols(4).rows(1).value([390.0, 400.0, 410.0, 420.0]),
                }),
                '[1]': lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda x: x.cols(4).rows(1).value([430.0, 440.0, 450.0, 460.0]),
                    1: lambda x: x.cols(4).rows(1).value([470.0, 480.0, 490.0, 500.0]),
                }),
            }),
            '[2]': lambda x: x.cols(0).rows(0).members({
                '[0]': lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda x: x.cols(4).rows(1).value([510.0, 520.0, 530.0, 540.0]),
                    1: lambda x: x.cols(4).rows(1).value([550.0, 560.0, 570.0, 580.0]),
                }),
                '[1]': lambda x: x.cols(0).rows(0).arraySize(2).members({
                    0: lambda x: x.cols(4).rows(1).value([590.0, 600.0, 610.0, 620.0]),
                    1: lambda x: x.cols(4).rows(1).value([630.0, 640.0, 650.0, 660.0]),
                }),
            }),
        })

        # struct vec3_1 { vec3 a; float b; };
        # struct nested { vec3_1 a; vec4 b[4]; vec3_1 c[4]; };
        # nested G[2];
        # Due to lacking reflection, we don't know that G[x].a.a exists, as we only reference G[n].a.b
        # Similarly we don't know that G[x].c[y].b exists
        var_check.check('G').cols(0).rows(0).arraySize(2).members({
            # G[0]
            0: lambda s: s.cols(0).rows(0).structSize(3).members({
                'a': lambda x: x.cols(0).rows(0).members({
                    'a': lambda y: y.cols(3).rows(1).value([680.0, 690.0, 700.0]),
                    'b': lambda y: y.cols(1).rows(1).value([710.0]),
                }),
                'b': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(4).rows(1),
                    1: lambda y: y.cols(4).rows(1),
                    2: lambda y: y.cols(4).rows(1),
                    3: lambda y: y.cols(4).rows(1),
                }),
                'c': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                    1: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                    2: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                    3: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                }),
            }),
            # G[1]
            1: lambda s: s.cols(0).rows(0).structSize(3).members({
                'a': lambda x: x.cols(0).rows(0).members({
                    'a': lambda y: y.cols(3).rows(1).value([1040.0, 1050.0, 1060.0]),
                    'b': lambda y: y.cols(1).rows(1).value([1070.0]),
                }),
                'b': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(4).rows(1).value([1080.0, 1090.0, 1100.0, 1110.0]),
                    1: lambda y: y.cols(4).rows(1).value([1120.0, 1130.0, 1140.0, 1150.0]),
                    2: lambda y: y.cols(4).rows(1).value([1160.0, 1170.0, 1180.0, 1190.0]),
                    3: lambda y: y.cols(4).rows(1).value([1200.0, 1210.0, 1220.0, 1230.0]),
                }),
                'c': lambda x: x.cols(0).rows(0).arraySize(4).members({
                    0: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                    1: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                    2: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                    3: lambda y: y.cols(0).rows(0).members({
                        'a': lambda z: z.cols(3).rows(1).value([1360.0, 1370.0, 1380.0]),
                        'b': lambda z: z.cols(1).rows(1),
                    }),
                }),
            }),
        })

        var_check.done()

        rdtest.log.success("Bare uniform variables are as expected")



import renderdoc as rd
import rdtest


class VK_Empty_Capture(rdtest.TestCase):
    demos_test_name = 'VK_Empty_Capture'
    demos_frame_cap = 100

    def check_capture(self):
        draws = self.controller.GetDrawcalls()

        self.check(len(draws) == 1)
        self.check('End' in draws[0].name)
        self.check(draws[0].eventId == 1)


import renderdoc as rd
import rdtest


class D3D12_List_Types(rdtest.TestCase):
    demos_test_name = 'D3D12_List_Types'

    def check_capture(self):
        draw = self.find_draw("Draw")

        self.controller.SetFrameEvent(draw.eventId, False)

        self.check_triangle(out=draw.outputs[0], fore=[0.0, 1.0, 1.0, 1.0])

        postvs_data = self.get_postvs(draw, rd.MeshDataStage.VSOut, 0, draw.numIndices)

        postvs_ref = {
            0: {
                'vtx': 0,
                'idx': 0,
                'SV_POSITION': [-0.5, -0.5, 0.0, 1.0],
                'COLOR': [0.0, 1.0, 1.0, 1.0],
                'TEXCOORD': [1234.0, 5678.0],
            },
            1: {
                'vtx': 1,
                'idx': 1,
                'SV_POSITION': [0.0, 0.5, 0.0, 1.0],
                'COLOR': [0.0, 1.0, 1.0, 1.0],
                'TEXCOORD': [1234.0, 5678.0],
            },
            2: {
                'vtx': 2,
                'idx': 2,
                'SV_POSITION': [0.5, -0.5, 0.0, 1.0],
                'COLOR': [0.0, 1.0, 1.0, 1.0],
                'TEXCOORD': [1234.0, 5678.0],
            },
        }

        self.check_mesh_data(postvs_ref, postvs_data)




import renderdoc as rd
import rdtest
import os

try:
    import tkinter
except ImportError as ex:
    tkinter = None

class D3D12_RGP_Capture(rdtest.TestCase):
    demos_test_name = 'D3D12_Simple_Triangle'

    def check_support(self):
        if tkinter is None:
            return False, 'tkinter is required but not available'
        
        return super().check_support()

    def check_capture(self):
        apiprops: rd.APIProperties = self.controller.GetAPIProperties()

        if not apiprops.rgpCapture:
            rdtest.log.print("RGP capture not tested")
            return


        # On D3D12 we need to create a real window
        window = tkinter.Tk()
        window.geometry("1280x720")

        path = self.controller.CreateRGPProfile(rd.CreateWin32WindowingData(int(window.frame(), 16)))

        rdtest.log.print("RGP capture created: '{}'".format(path))

        if os.path.exists(path) and os.path.getsize(path) > 100:
            rdtest.log.success("RGP capture created successfully")
        else:
            raise rdtest.TestFailureException("RGP capture failed")




import renderdoc as rd
import rdtest
from typing import List

def value_selector(x): return x.floatValue
def passed(x): return x.Passed()
def event_id(x): return x.eventId
def culled(x): return x.backfaceCulled
def depth_test_failed(x): return x.depthTestFailed
def depth_clipped(x): return x.depthClipped
def depth_bounds_failed(x): return x.depthBoundsFailed
def scissor_clipped(x): return x.scissorClipped
def stencil_test_failed(x): return x.stencilTestFailed
def shader_discarded(x): return x.shaderDiscarded
def shader_out_col(x): return value_selector(x.shaderOut.col)
def shader_out_depth(x): return x.shaderOut.depth
def pre_mod_col(x): return value_selector(x.preMod.col)
def post_mod_col(x): return value_selector(x.postMod.col)
def shader_out_depth(x): return x.shaderOut.depth
def pre_mod_depth(x): return x.preMod.depth
def post_mod_depth(x): return x.postMod.depth
def primitive_id(x): return x.primitiveID
def unboundPS(x): return x.unboundPS

class VK_Pixel_History(rdtest.TestCase):
    demos_test_name = 'VK_Pixel_History'
    demos_frame_cap = 5

    def check_capture(self):
        apiprops: rd.APIProperties = self.controller.GetAPIProperties()

        if not apiprops.pixelHistory:
            rdtest.log.print("Vulkan pixel history not tested")
            return

        self.is_depth = False
        self.primary_test()
        self.multisampled_image_test()
        self.secondary_cmd_test()

        self.is_depth = True
        self.depth_target_test()
        self.multisampled_image_test()

    def primary_test(self):
        test_marker: rd.DrawcallDescription = self.find_draw("Test Begin")
        self.controller.SetFrameEvent(test_marker.next.eventId, True)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        rt: rd.BoundResource = pipe.GetOutputTargets()[0]

        tex = rt.resourceId
        tex_details = self.get_texture(tex)

        sub = rd.Subresource()
        if tex_details.arraysize > 1:
            sub.slice = rt.firstSlice
        if tex_details.mips > 1:
            sub.mip = rt.firstMip

        begin_renderpass_eid = self.find_draw("Begin RenderPass").next.eventId
        depth_write_eid = self.find_draw("Depth Write").next.eventId
        stencil_write_eid = self.find_draw("Stencil Write").next.eventId
        unbound_fs_eid = self.find_draw("Unbound Fragment Shader").next.eventId
        background_eid = self.find_draw("Background").next.eventId
        cull_eid = self.find_draw("Cull Front").next.eventId
        test_eid = self.find_draw("Test Begin").next.eventId
        fixed_scissor_fail_eid = self.find_draw("Fixed Scissor Fail").next.eventId
        fixed_scissor_pass_eid = self.find_draw("Fixed Scissor Pass").next.eventId
        dynamic_stencil_ref_eid = self.find_draw("Dynamic Stencil Ref").next.eventId
        dynamic_stencil_mask_eid = self.find_draw("Dynamic Stencil Mask").next.eventId
        depth_test_eid = self.find_draw("Depth Test").next.eventId
        depth_bounds_prep_eid = self.find_draw("Depth Bounds Prep").next.eventId
        depth_bounds_clip_eid = self.find_draw("Depth Bounds Clip").next.eventId

        # For pixel 190, 149 inside the red triangle
        x, y = 190, 149
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, unbound_fs_eid], [passed, True], [unboundPS, True], [primitive_id, 0]],
            [[event_id, stencil_write_eid], [passed, True]],
            [[event_id, background_eid], [depth_test_failed, True], [post_mod_col, (1.0, 0.0, 0.0, 1.0)]],
            [[event_id, test_eid], [stencil_test_failed, True]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 190, 150
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, depth_write_eid], [passed, True]],
            [[event_id, background_eid], [depth_test_failed, True]],
            [[event_id, cull_eid], [culled, True]],
            [[event_id, test_eid], [depth_test_failed, True]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 200, 50
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, background_eid], [passed, True]],
            [[event_id, test_eid], [passed, True], [primitive_id, 7]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 150, 250
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, background_eid], [shader_discarded, True]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 330, 145
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, test_eid], [passed, True], [primitive_id, 3], [shader_out_col, (0.0, 0.0, 0.0, 2.75)]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 340, 145
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, test_eid], [passed, False], [depth_clipped, True]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 330, 105
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, depth_bounds_prep_eid], [passed, True], [primitive_id, 0], [shader_out_col, (1.0, 0.0, 0.0, 2.75)]],
            [[event_id, depth_bounds_clip_eid], [passed, True], [primitive_id, 0], [shader_out_col, (0.0, 1.0, 0.0, 2.75)]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 320, 105
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, depth_bounds_prep_eid], [passed, True], [primitive_id, 0], [shader_out_col, (1.0, 0.0, 0.0, 2.75)]],
            [[event_id, depth_bounds_clip_eid], [passed, False], [depth_bounds_failed, True]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        x, y = 345, 105
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, depth_bounds_prep_eid], [passed, True], [primitive_id, 0], [shader_out_col, (1.0, 0.0, 0.0, 2.75)]],
            [[event_id, depth_bounds_clip_eid], [passed, False], [depth_bounds_failed, True]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        rdtest.log.print("Testing dynamic state pipelines")
        self.controller.SetFrameEvent(dynamic_stencil_mask_eid, True)

        x, y = 100, 250
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, background_eid], [passed, True]],
            [[event_id, fixed_scissor_fail_eid], [scissor_clipped, True]],
            [[event_id, fixed_scissor_pass_eid], [passed, True], [shader_out_col, (0.0, 1.0, 0.0, 2.75)],
             [post_mod_col, (0.0, 1.0, 0.0, 1.0)]],
            [[event_id, dynamic_stencil_ref_eid], [passed, True], [shader_out_col, (0.0, 0.0, 1.0, 2.75)],
             [post_mod_col, (0.0, 0.0, 1.0, 1.0)]],
            [[event_id, dynamic_stencil_mask_eid], [passed, True], [shader_out_col, (0.0, 1.0, 1.0, 2.75)],
             [post_mod_col, (0.0, 1.0, 1.0, 1.0)]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        rdtest.log.print("Testing depth test for per fragment reporting")
        self.controller.SetFrameEvent(depth_test_eid, True)

        x, y = 275, 260
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True]],
            [[event_id, background_eid], [passed, True]],
            [[event_id, depth_test_eid], [primitive_id, 0], [depth_test_failed, True],
             [shader_out_col, (1.0, 1.0, 1.0, 2.75)], [shader_out_depth, 0.97], [post_mod_col, (1.0, 0.0, 1.0, 1.0)],
             [post_mod_depth, 0.95]],
            [[event_id, depth_test_eid], [primitive_id, 1], [depth_test_failed, False],
             [shader_out_col, (1.0, 1.0, 0.0, 2.75)], [shader_out_depth, 0.20], [post_mod_col, (1.0, 1.0, 0.0, 1.0)],
             [post_mod_depth, 0.20]],
            [[event_id, depth_test_eid], [primitive_id, 2], [depth_test_failed, True],
             [shader_out_col, (1.0, 0.0, 0.0, 2.75)], [shader_out_depth, 0.30], [post_mod_col, (1.0, 1.0, 0.0, 1.0)],
             [post_mod_depth, 0.20]],
            [[event_id, depth_test_eid], [primitive_id, 3], [depth_test_failed, False],
             [shader_out_col, (0.0, 0.0, 1.0, 2.75)], [shader_out_depth, 0.10], [post_mod_col, (0.0, 0.0, 1.0, 1.0)],
             [post_mod_depth, 0.10]],
            [[event_id, depth_test_eid], [primitive_id, 4], [depth_test_failed, False],
             [shader_out_col, (1.0, 1.0, 1.0, 2.75)], [shader_out_depth, 0.05], [post_mod_col, (0.0, 0.0, 1.0, 1.0)],
             [post_mod_depth, 0.10]],
        ]
        self.check_events(events, modifs, False)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

    def multisampled_image_test(self):
        test_marker: rd.DrawcallDescription = self.find_draw("Multisampled: test")
        draw_eid = test_marker.next.eventId
        self.controller.SetFrameEvent(draw_eid, True)

        pipe: rd.PipeState = self.controller.GetPipelineState()
        rt: rd.BoundResource = pipe.GetOutputTargets()[0]

        if self.is_depth:
            rt: rd.BoundResource = pipe.GetDepthTarget()

        sub = rd.Subresource()
        tex = rt.resourceId
        tex_details = self.get_texture(tex)
        if tex_details.arraysize > 1:
            sub.slice = rt.firstSlice

        beg_renderpass_eid = self.find_draw("Multisampled: begin renderpass").next.eventId

        x, y = 140, 130
        sub.sample = 1
        rdtest.log.print("Testing pixel {}, {} at sample {}".format(x, y, sub.sample))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)

        events = [
            [[event_id, beg_renderpass_eid], [passed, True], [post_mod_depth, 0.0]],
            [[event_id, draw_eid], [passed, True], [primitive_id, 0], [pre_mod_depth, 0.0], [shader_out_depth, 0.9],
             [post_mod_depth, 0.9]],
            [[event_id, draw_eid], [passed, True], [primitive_id, 1], [shader_out_depth, 0.95], [post_mod_depth, 0.95]],
        ]

        if not self.is_depth:
            events[0] += [[post_mod_col, (0.0, 1.0, 0.0, 1.0)]]
            events[1] += [[shader_out_col, (1.0, 0.0, 1.0, 2.75)], [post_mod_col, (1.0, 0.0, 1.0, 1.0)]]
            events[2] += [[shader_out_col, (0.0, 0.0, 1.0, 2.75)], [post_mod_col, (0.0, 0.0, 1.0, 1.0)]]

        self.check_events(events, modifs, True)

        if self.is_depth:
            self.check_pixel_value(tex, x, y, [modifs[-1].postMod.depth, float(modifs[-1].postMod.stencil)/255.0, 0.0, 1.0], sub=sub, cast=rt.typeCast)
        else:
            self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        sub.sample = 2
        rdtest.log.print("Testing pixel {}, {} at sample {}".format(x, y, sub.sample))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, beg_renderpass_eid], [passed, True], [post_mod_depth, 0.0]],
            [[event_id, draw_eid], [passed, True], [primitive_id, 0], [pre_mod_depth, 0.0], [shader_out_depth, 0.9],
             [post_mod_depth, 0.9]],
            [[event_id, draw_eid], [passed, True], [primitive_id, 1], [shader_out_depth, 0.95], [post_mod_depth, 0.95]],
        ]

        if not self.is_depth:
            events[0] += [[post_mod_col, (0.0, 1.0, 0.0, 1.0)]]
            events[1] += [[shader_out_col, (1.0, 0.0, 1.0, 2.75)], [post_mod_col, (1.0, 0.0, 1.0, 1.0)]]
            events[2] += [[shader_out_col, (0.0, 1.0, 1.0, 2.75)], [post_mod_col, (0.0, 1.0, 1.0, 1.0)]]

        self.check_events(events, modifs, True)

        if self.is_depth:
            self.check_pixel_value(tex, x, y,
                                   [modifs[-1].postMod.depth, float(modifs[-1].postMod.stencil) / 255.0, 0.0, 1.0],
                                   sub=sub, cast=rt.typeCast)
        else:
            self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

    def secondary_cmd_test(self):
        secondary_marker: rd.DrawcallDescription = self.find_draw("Secondary: red and blue")
        self.controller.SetFrameEvent(secondary_marker.next.eventId, True)

        pipe: rd.PipeState = self.controller.GetPipelineState()
        rt: rd.BoundResource = pipe.GetOutputTargets()[0]
        sub = rd.Subresource()
        tex = rt.resourceId
        tex_details = self.get_texture(tex)
        if tex_details.arraysize > 1:
            sub.slice = rt.firstSlice
        if tex_details.mips > 1:
            sub.mip = rt.firstMip

        sec_beg_renderpass_eid = self.find_draw("Begin RenderPass Secondary").next.eventId
        background_eid = self.find_draw("Secondary: background").next.eventId
        culled_eid = self.find_draw("Secondary: culled").next.eventId
        sec_red_and_blue = self.find_draw("Secondary: red and blue").next.eventId

        # Test culling
        x, y = 70, 40
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, sec_beg_renderpass_eid], [passed, True], [post_mod_col, (0.0, 1.0, 0.0, 1.0)]],
            [[event_id, background_eid], [passed, True], [pre_mod_col, (0.0, 1.0, 0.0, 1.0)]],
            [[event_id, culled_eid], [passed, False], [culled, True]],
        ]
        self.check_events(events, modifs, True)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        # Blue triangle
        x, y = 40, 40
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, sec_beg_renderpass_eid], [passed, True], [post_mod_col, (0.0, 1.0, 0.0, 1.0)]],
            # This is the first event in the command buffer, should have pre-mod
            [[event_id, background_eid], [passed, True], [pre_mod_col, (0.0, 1.0, 0.0, 1.0)]],
            # This is the last event in the command buffer, should have post-mod
            [[event_id, sec_red_and_blue], [passed, True], [post_mod_col, (0.0, 0.0, 1.0, 1.0)]],
        ]
        self.check_events(events, modifs, True)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

        # Didn't get post mod for background_eid
        self.controller.SetFrameEvent(background_eid, True)
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, sec_beg_renderpass_eid]],
            # The only event, should have both pre and post mod.
            [[event_id, background_eid], [passed, True], [pre_mod_col, (0.0, 1.0, 0.0, 1.0)], [post_mod_col, (1.0, 0.0, 1.0, 1.0)]],
        ]
        self.check_events(events, modifs, True)
        self.check_pixel_value(tex, x, y, value_selector(modifs[-1].postMod.col), sub=sub, cast=rt.typeCast)

    def depth_target_test(self):
        test_marker: rd.DrawcallDescription = self.find_draw("Test Begin")
        self.controller.SetFrameEvent(test_marker.next.eventId, True)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        rt: rd.BoundResource = pipe.GetDepthTarget()

        tex = rt.resourceId
        tex_details = self.get_texture(tex)

        sub = rd.Subresource()
        if tex_details.arraysize > 1:
            sub.slice = rt.firstSlice
        if tex_details.mips > 1:
            sub.mip = rt.firstMip

        begin_renderpass_eid = self.find_draw("Begin RenderPass").next.eventId
        background_eid = self.find_draw("Background").next.eventId
        test_eid = self.find_draw("Test Begin").next.eventId

        x, y = 200, 190
        rdtest.log.print("Testing pixel {}, {}".format(x, y))
        modifs: List[rd.PixelModification] = self.controller.PixelHistory(tex, x, y, sub, rt.typeCast)
        events = [
            [[event_id, begin_renderpass_eid], [passed, True], [post_mod_depth, 1.0]],
            [[event_id, background_eid], [passed, True], [primitive_id, 0], [pre_mod_depth, 1.0], [post_mod_depth, 0.95]],
            [[event_id, test_eid], [passed, True], [depth_test_failed, False], [primitive_id, 0], [shader_out_depth, 0.5], [post_mod_depth, 0.5]],
            [[event_id, test_eid], [passed, False], [depth_test_failed, True], [primitive_id, 1], [shader_out_depth, 0.6], [post_mod_depth, 0.5]],
        ]
        self.check_events(events, modifs, False)


    def check_events(self, events, modifs, hasSecondary):
        self.check(len(modifs) == len(events), "Expected {} events, got {}".format(len(events), len(modifs)))
        # Check for consistency first. For secondary command buffers,
        # might not have all information, so don't check for consistency
        if not hasSecondary:
            self.check_modifs_consistent(modifs)
        for i in range(len(modifs)):
            for c in range(len(events[i])):
                expected = events[i][c][1]
                actual = events[i][c][0](modifs[i])
                if not rdtest.value_compare(actual, expected, eps=1.0/256.0):
                    raise rdtest.TestFailureException(
                        "eventId {}, primitiveID {}: testing {} expected {}, got {}".format(modifs[i].eventId, modifs[i].primitiveID,
                                                                            events[i][c][0].__name__,
                                                                            expected,
                                                                            actual))

    def check_modifs_consistent(self, modifs):
        # postmod of each should match premod of the next
        for i in range(len(modifs) - 1):
            a = value_selector(modifs[i].postMod.col)
            b = value_selector(modifs[i + 1].preMod.col)

            if self.is_depth:
                a = (modifs[i].postMod.depth, modifs[i].postMod.stencil)
                b = (modifs[i + 1].preMod.depth, modifs[i + 1].preMod.stencil)

            if a != b:
                raise rdtest.TestFailureException(
                    "postmod at {} primitive {}: {} doesn't match premod at {} primitive {}: {}".format(modifs[i].eventId,
                                                                              modifs[i].primitiveID,
                                                                              a,
                                                                              modifs[i + 1].eventId,
                                                                              modifs[i + 1].primitiveID,
                                                                              b))

        # Check that if the test failed, its postmod is the same as premod
        for i in range(len(modifs)):
            if not modifs[i].Passed():
                a = value_selector(modifs[i].preMod.col)
                b = value_selector(modifs[i].postMod.col)

                if self.is_depth:
                    a = (modifs[i].preMod.depth, modifs[i].preMod.stencil)
                    b = (modifs[i].postMod.depth, modifs[i].postMod.stencil)

                if not rdtest.value_compare(a, b):
                    raise rdtest.TestFailureException(
                        "postmod at {} primitive {}: {} doesn't match premod: {}".format(modifs[i].eventId,
                                                                                         modifs[i].primitiveID, b, a))



import rdtest
import struct
import renderdoc as rd


class D3D11_Stream_Out(rdtest.TestCase):
    demos_test_name = 'D3D11_Stream_Out'

    def check_capture(self):
        draw = self.find_draw("Draw")

        self.check(draw is not None)

        self.controller.SetFrameEvent(draw.eventId, False)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        # Get the input data as our reference

        # First draw should have data at offset 0. First buffer has positions, second has colors (with doubled stride)

        vsin = self.get_vsin(draw)

        pos = [(*v['POSITION'], 1.0) for v in vsin]
        col = [v['COLOR'] for v in vsin]

        d3d11pipe: rd.D3D11State = self.controller.GetD3D11PipelineState()

        so: rd.D3D11StreamOut = d3d11pipe.streamOut
        so_bytes = self.controller.GetBufferData(so.outputs[0].resourceId, so.outputs[0].byteOffset, 0)

        for i,p in enumerate(pos):
            so_p = struct.unpack_from("4f", so_bytes, 0 + 4*4*i)
            if not rdtest.value_compare(p, so_p):
                raise rdtest.TestFailureException("Streamed-out position {} doesn't match expected {}".format(so_p, p))

        so_bytes = self.controller.GetBufferData(so.outputs[1].resourceId, so.outputs[1].byteOffset, 0)

        for i,c in enumerate(col):
            so_c = struct.unpack_from("4f", so_bytes, 0 + 8*4*i)
            if not rdtest.value_compare(c, so_c):
                raise rdtest.TestFailureException("Streamed-out color {} doesn't match expected {}".format(so_c, c))

        draw_auto = self.find_draw("DrawAuto", draw.eventId)

        # First draw should be 3 vertices
        if not rdtest.value_compare(draw_auto.numIndices, 3):
            raise rdtest.TestFailureException("First DrawAuto() draws {} vertices".format(draw_auto.numIndices))

        draw_auto = self.find_draw("DrawAuto", draw_auto.eventId+1)

        # Second draw should be 6 vertices (3 vertices, instanced twice
        if not rdtest.value_compare(draw_auto.numIndices, 6):
            raise rdtest.TestFailureException("Second DrawAuto() draws {} vertices".format(draw_auto.numIndices))
        if not rdtest.value_compare(draw_auto.numInstances, 1):
            raise rdtest.TestFailureException("Second DrawAuto() draws {} instances".format(draw_auto.numInstances))

        rdtest.log.success("First draw stream-out data is correct")



import renderdoc as rd
import struct
import rdtest


class VK_Misaligned_Dirty(rdtest.TestCase):
    demos_test_name = 'VK_Misaligned_Dirty'

    def get_capture_options(self):
        opts = rdtest.TestCase.get_capture_options(self)
        opts.apiValidation = True
        return opts

    def get_replay_options(self):
        opts = rdtest.TestCase.get_replay_options(self)
        # Set a balanced optimisation level to ensure that written ranges are cleared instead of being either restored
        # or ignored
        opts.optimisation = rd.ReplayOptimisationLevel.Balanced
        return opts

    def check_capture(self):
        draw = self.find_draw("Draw")

        self.check(draw is not None)

        self.controller.SetFrameEvent(draw.eventId, False)

        self.check(len(self.controller.GetFrameInfo().debugMessages) == 0)
        self.check(len(self.controller.GetDebugMessages()) == 0)

        rdtest.log.success("No debug messages found")

        pipe: rd.PipeState = self.controller.GetPipelineState()

        postvs_data = self.get_postvs(draw, rd.MeshDataStage.VSOut, 0, draw.numIndices)

        val = 2.0 / 3.0

        postvs_ref = {
            0: {
                'vtx': 0,
                'idx': 0,
                'gl_PerVertex_var.gl_Position': [-val, val, val, 1.0],
                'vertOut.col': [0.0, 1.0, 0.0, 1.0],
            },
            1: {
                'vtx': 1,
                'idx': 1,
                'gl_PerVertex_var.gl_Position': [0.0, -val, val, 1.0],
                'vertOut.col': [0.0, 1.0, 0.0, 1.0],
            },
            2: {
                'vtx': 2,
                'idx': 2,
                'gl_PerVertex_var.gl_Position': [val, val, val, 1.0],
                'vertOut.col': [0.0, 1.0, 0.0, 1.0],
            },
        }

        self.check_mesh_data(postvs_ref, postvs_data)

        rdtest.log.success("vertex output is as expected")

        tex = pipe.GetOutputTargets()[0].resourceId

        texdetails = self.get_texture(tex)

        coords = [
            [int(texdetails.width * 1 / 3) + 5, int(texdetails.height * 2 / 3) - 5],
            [int(texdetails.width * 1 / 2) + 0, int(texdetails.height * 1 / 3) + 5],
            [int(texdetails.width * 2 / 3) + 5, int(texdetails.height * 2 / 3) - 5],
        ]

        for coord in coords:
            self.check_pixel_value(tex, coord[0], coord[1], [0.0, 1.0, 0.0, 1.0])

        rdtest.log.success("picked values are as expected")

        checkpoint1 = self.find_draw("First Submit")
        checkpoint2 = self.find_draw("Second Submit")
        checkpoint3 = self.find_draw("Third Submit")

        self.check(checkpoint1 is not None)
        self.check(checkpoint2 is not None)
        self.check(checkpoint3 is not None)

        resources = self.controller.GetResources()

        copy_src = None
        vb = None

        for r in resources:
            if r.name == 'copy_src':
                copy_src = r.resourceId
            elif r.name == 'vb':
                vb = r.resourceId

        self.check(copy_src is not None)
        self.check(vb is not None)

        self.controller.SetFrameEvent(checkpoint1.eventId, False)

        val = struct.unpack('f', self.controller.GetBufferData(copy_src, 116, 4))
        self.check(val[0] == 11.0)

        self.controller.SetFrameEvent(checkpoint2.eventId, False)

        val = struct.unpack('f', self.controller.GetBufferData(copy_src, 116, 4))
        self.check(val[0] == 12.0)
        val = struct.unpack('f', self.controller.GetBufferData(vb, 116, 4))
        self.check(val[0] == 12.0)

        self.controller.SetFrameEvent(checkpoint3.eventId, False)

        val = struct.unpack('f', self.controller.GetBufferData(copy_src, 116, 4))
        self.check(val[0] == 11.0)

        rdtest.log.success("buffers have correct values in both submits")



#!/usr/bin/env python3

import os
import io
import sys
import re
import argparse
import json
import xml.etree.ElementTree as ET

# on msys, use crlf output
nl = None
if sys.platform == 'msys':
    nl = "\r\n"
    
# Get the file, relative to this script's location (same directory)
# that way we're not sensitive to CWD
pathname = os.path.abspath(os.path.dirname(sys.argv[0])) + os.path.sep

with open(pathname + 'spirv.core.grammar.json', mode='r') as f:
    spirv = json.load(f)

with open(pathname + 'extinst.glsl.std.450.grammar.json', mode='r') as f:
    glsl450 = json.load(f)

# open XML registry
registry = ET.parse(pathname + 'spir-v.xml').getroot()

# open the file for write
header = open(pathname + 'spirv_gen.h', mode='w', newline = nl)
ops_header = open(pathname + 'spirv_op_helpers.h', mode='w', newline = nl)
cpp = open(pathname + 'spirv_gen.cpp', mode='w', newline = nl)

###############################################################################
##
## Headers
##
###############################################################################

def prefix_star(line):
    if line == '':
        return ' *'
    else:
        return ' * ' + line

def operand_name(name, lowercase_first = True):
    name = name.replace('\n', ' ')
    # special case a few very awkward names
    if re.search(r'member [0-9].*\.\.\.', name, re.RegexFlag.I):
        return 'members'
    if re.search(r'parameter [0-9].*\.\.\.', name, re.RegexFlag.I):
        return 'parameters'
    if re.search(r'argument [0-9].*\.\.\.', name, re.RegexFlag.I):
        return 'arguments'
    if re.search(r'variable, parent.*\.\.\.', name, re.RegexFlag.I):
        return 'parents'

    name = re.sub(r'<<(.*),(.*)>>', r'\2', name)
    name = re.sub(r'[ \'~<>./-]', '', name)

    if name.lower() == 'interface':
        return 'iface'

    if name.lower() == 'default':
        return 'def'

    if lowercase_first:
        return name[0].lower() + name[1:]
    else:
        return name

copyright = '''
/******************************************************************************
 * The MIT License (MIT)
 *
 * Copyright (c) 2019-2021 Baldur Karlsson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 ******************************************************************************/

/******************************************************************************
 * Generated from Khronos SPIR-V machine-readable JSON grammar.
 *
{}
 ******************************************************************************/
'''.format("\n".join([prefix_star(line.strip()) for line in spirv['copyright']])).strip()

header.write('''{copyright}

#pragma once

// This file is autogenerated with gen_spirv_code.py - any changes will be overwritten next time
// that script is run.
// $ ./gen_spirv_code.py

// We need to disable clang-format since this file is programmatically generated
// clang-format off

#include <stdint.h>
#include "api/replay/apidefs.h"
#include "api/replay/stringise.h"

#undef None
#undef CopyMemory
#undef MemoryBarrier

namespace rdcspv
{{
static const uint32_t MagicNumber = {magic};
static const uint32_t VersionMajor = {major};
static const uint32_t VersionMinor = {minor};
static const uint32_t VersionRevision = {revision};
static const uint32_t VersionPacked = ({major} << 16) | ({minor} << 8);
static const uint32_t OpCodeMask = 0xffff;
static const uint32_t WordCountShift = 16;
static const uint32_t FirstRealWord = 5;

struct Id
{{
  constexpr inline Id() : id(0) {{}}
  // only allow explicit functions to cast to/from uint32_t
  constexpr static inline Id fromWord(uint32_t i) {{ return Id(i); }}
  inline uint32_t value() const {{ return id; }}
  constexpr inline explicit operator bool() const {{ return id != 0; }}
  constexpr inline bool operator==(const Id o) const {{ return id == o.id; }}
  constexpr inline bool operator!=(const Id o) const {{ return id != o.id; }}
  constexpr inline bool operator<(const Id o) const {{ return id < o.id; }}
  constexpr inline bool operator==(const uint32_t o) const {{ return id == o; }}
  constexpr inline bool operator!=(const uint32_t o) const {{ return id != o; }}
  constexpr inline bool operator<(const uint32_t o) const {{ return id < o; }}
private:
  constexpr inline Id(uint32_t i) : id(i) {{}}
  uint32_t id;
}};

enum class Generator : uint32_t
{{'''.format(copyright = copyright, magic = spirv['magic_number'], major = spirv['major_version'], minor = spirv['minor_version'], revision = spirv['revision']))

generator_tostr = ''

for gen in registry.findall('ids[@type=\'vendor\']/id[@tool]'):
    name = operand_name(gen.attrib['tool'], lowercase_first=False)

    tostr = '{} from {} - {}'.format(gen.attrib['tool'], gen.attrib['vendor'], gen.attrib['comment'])

    generator_tostr += '    STRINGISE_ENUM_CLASS_NAMED({}, "{}");\n'.format(name, tostr.replace('"', '\\"').replace('\\', '\\\\'))
    header.write('\n  {} = {},'.format(name, gen.attrib['value']))

header.write('\n};\n\n')

ops_header.write('''{copyright}

#pragma once

// This file is autogenerated with gen_spirv_code.py - any changes will be overwritten next time
// that script is run.
// $ ./gen_spirv_code.py

// We need to disable clang-format since this file is programmatically generated
// clang-format off

#include <functional>
#include <set>
#include <stdint.h>
#include "api/replay/apidefs.h"
#include "api/replay/rdcstr.h"
#include "api/replay/rdcarray.h"
#include "api/replay/stringise.h"

#undef None
#undef CopyMemory
#undef MemoryBarrier

#include "spirv_common.h"
#include "spirv_gen.h"

namespace rdcspv
{{

template<typename Type>
Type DecodeParam(const ConstIter &it, uint32_t &word);

template<>
inline uint32_t DecodeParam(const ConstIter &it, uint32_t &word)
{{
  if(word >= it.size()) return 0;
  
  uint32_t ret = it.word(word);
  word += 1;
  return ret;
}}

template<>
inline Id DecodeParam<Id>(const ConstIter &it, uint32_t &word)
{{
  if(word >= it.size()) return Id();
  
  Id ret = Id::fromWord(it.word(word));
  word += 1;
  return ret;
}}

template<>
inline rdcstr DecodeParam<rdcstr>(const ConstIter &it, uint32_t &word)
{{
  if(word >= it.size()) return "";
  
  rdcstr ret = (const char *)&it.word(word);
  word += uint32_t(ret.size() / 4) + 1;
  return ret;
}}

template<typename Type>
rdcarray<Type> MultiParam(const ConstIter &it, uint32_t &word)
{{
  rdcarray<Type> ret;
  while(word < it.size())
  {{
    Type t = DecodeParam<Type>(it, word);
    ret.push_back(t);
  }}
  return ret;
}}

inline void EncodeParam(rdcarray<uint32_t> &words, const rdcstr &str)
{{
  size_t i=0, remainingChars = str.size() + 1;
  while(remainingChars > 0)
  {{
    uint32_t word = 0;
    for(size_t w=0; w < remainingChars && w < 4; w++)
      word |= uint32_t(str[i+w]) << (w*8);
    words.push_back(word);
    
    i += 4;
    if(remainingChars < 4)
      remainingChars = 0;
    else
      remainingChars -= 4;
  }}
}}

'''.format(copyright = copyright))

cpp.write('''{copyright}

// This file is autogenerated with gen_spirv_code.py - any changes will be overwritten next time
// that script is run.
// $ ./gen_spirv_code.py

// We need to disable clang-format since this file is programmatically generated
// clang-format off

#include "spirv_gen.h"
#include "os/os_specific.h"
#include "common/formatting.h"
#include "spirv_op_helpers.h"

'''.format(copyright = copyright))

###############################################################################
##
## Operands (declare enums, stringise, preprocess)
##
###############################################################################

positional_names = [ 'first', 'second', 'third' ]
kinds = {}

for operand_kind in spirv['operand_kinds']:
    name = operand_kind['kind']

    if 'enumerants' in operand_kind:
        operand_kind['has_params'] = any(['parameters' in value for value in operand_kind['enumerants']])
    else:
        operand_kind['has_params'] = False

    kinds[name] = operand_kind

    operand_kind['push_words'] = lambda name: 'words.push_back((uint32_t){});'.format(name)
    operand_kind['from_words'] = None
    operand_kind['is_id'] = False

    if operand_kind['category'] == 'ValueEnum':
        operand_kind['size'] = 1
        operand_kind['def_name'] = name[0].lower() + name[1:]
        operand_kind['def_value'] = name + '::Invalid'
        operand_kind['type'] = name

        decl = ''
        stringise = ''

        used = []

        for value in operand_kind['enumerants']:
            value_name = value['enumerant']
            if value_name[0].isdigit():
                value_name = '_' + value_name
            decl += '  {} = {},\n'.format(value_name, value['value'])

            if value['value'] in used:
                continue

            used.append(value['value'])

            if value_name != value['enumerant']:
                stringise += '    STRINGISE_ENUM_CLASS_NAMED({}, "{}");\n'.format(value_name, value['enumerant'])
            else:
                stringise += '    STRINGISE_ENUM_CLASS({});\n'.format(value_name)

        header.write('''enum class {name} : uint32_t
{{
{values}
  Max,
  Invalid = ~0U,
}};

'''.format(name = name, values = decl.rstrip()))

        cpp.write('''template <>
rdcstr DoStringise(const rdcspv::{name} &el)
{{
  BEGIN_ENUM_STRINGISE(rdcspv::{name});
  {{
{values}
  }}
  END_ENUM_STRINGISE();
}}

'''.format(name = name, values = stringise.rstrip()))
    elif operand_kind['category'] == 'BitEnum':
        operand_kind['size'] = 1
        operand_kind['def_name'] = name[0].lower() + name[1:]
        operand_kind['def_value'] = name + '::None'
        operand_kind['type'] = name

        used = []

        decl = ''
        stringise = ''
        for value in operand_kind['enumerants']:
            decl += '  {} = {},\n'.format(value['enumerant'], value['value'])

            if value['value'] in used:
                continue

            used.append(value['value'])

            if value['enumerant'] == 'None':
                stringise += '    STRINGISE_BITFIELD_CLASS_VALUE(None);\n\n'
            else:
                stringise += '    STRINGISE_BITFIELD_CLASS_BIT({});\n'.format(value['enumerant'])

        header.write('''enum class {name} : uint32_t
{{
{values}
  Max,
  Invalid = ~0U,
}};

BITMASK_OPERATORS({name});

'''.format(name = name, values = decl.rstrip()))

        cpp.write('''template <>
rdcstr DoStringise(const rdcspv::{name} &el)
{{
  BEGIN_BITFIELD_STRINGISE(rdcspv::{name});
  {{
{values}
  }}
  END_BITFIELD_STRINGISE();
}}

'''.format(name = name, values = stringise.rstrip()))
    # Hardcoded special types that we hardcode behaviour for
    elif (operand_kind['kind'] == 'IdRef'):
        operand_kind['size'] = 1
        operand_kind['def_name'] = 'id'
        operand_kind['def_value'] = 'Id()'
        operand_kind['type'] = 'Id'
        operand_kind['is_id'] = True
        operand_kind['push_words'] = lambda name: 'words.push_back({}.value());'.format(name)
        operand_kind['from_words'] = lambda name: 'Id::fromWord({})'.format(name)
    elif (operand_kind['kind'] == 'IdResultType' or
          operand_kind['kind'] == 'IdResult' or
          operand_kind['kind'] == 'IdMemorySemantics' or
          operand_kind['kind'] == 'IdScope'):
        operand_kind['size'] = 1
        operand_kind['type'] = name
        operand_kind['is_id'] = True
        operand_kind['def_name'] = name[2].lower() + name[3:]
        operand_kind['def_value'] = name + '()'
        operand_kind['push_words'] = lambda name: 'words.push_back({}.value());'.format(name)
        operand_kind['from_words'] = lambda name: 'Id::fromWord({})'.format(name)
        header.write('using {} = Id;\n\n'.format(name))
    # For simplicity, assume literal integers are 32-bit in size
    elif (operand_kind['kind'] == 'LiteralInteger'):
        operand_kind['size'] = 1
        operand_kind['def_name'] = 'num'
        operand_kind['def_value'] = '0'
        operand_kind['type'] = 'uint32_t'
    elif (operand_kind['kind'] == 'LiteralString'):
        operand_kind['size'] = -1000000
        operand_kind['type'] = 'rdcstr'
        operand_kind['def_name'] = 'str'
        operand_kind['def_value'] = '""'
        operand_kind['push_words'] = lambda name: 'EncodeParam(words, {});'.format(name)
        operand_kind['from_words'] = lambda name: 'DecodeParam({})'.format(name)
    elif (operand_kind['kind'] == 'LiteralContextDependentNumber' or
          operand_kind['kind'] == 'LiteralExtInstInteger' or
          operand_kind['kind'] == 'LiteralSpecConstantOpInteger'):
        operand_kind['size'] = None
    elif (operand_kind['kind'] == 'PairLiteralIntegerIdRef'):
        operand_kind['size'] = 2
        operand_kind['def_name'] = name[0].lower() + name[1:]
        operand_kind['def_value'] = '{0, Id()}'
        operand_kind['type'] = name
        operand_kind['push_words'] = lambda name: 'words.push_back((uint32_t){0}.first); words.push_back({0}.second.value());'.format(name)
        ops_header.write('struct {} {{ uint32_t first; Id second; }};\n\n'.format(name))
    elif (operand_kind['kind'] == 'PairIdRefLiteralInteger'):
        operand_kind['size'] = 2
        operand_kind['def_name'] = name[0].lower() + name[1:]
        operand_kind['def_value'] = '{Id(), 0}'
        operand_kind['type'] = name
        operand_kind['push_words'] = lambda name: 'words.push_back({0}.first.value()); words.push_back((uint32_t){0}.second);'.format(name)
        ops_header.write('struct {} {{ Id first; uint32_t second; }};\n\n'.format(name))
    elif (operand_kind['kind'] == 'PairIdRefIdRef'):
        operand_kind['size'] = 2
        operand_kind['def_name'] = name[0].lower() + name[1:]
        operand_kind['def_value'] = '{Id(), Id()}'
        operand_kind['type'] = name
        operand_kind['push_words'] = lambda name: 'words.push_back({0}.first.value()); words.push_back({0}.second.value());'.format(name)
        ops_header.write('struct {} {{ Id first, second; }};\n\n'.format(name))
        continue
    else:
        raise TypeError("Unexpected operand {} of type {}".format(operand_kind['kind'], operand_kind['category']))

    if operand_kind['from_words'] is None:
        operand_kind['from_words'] = lambda name,kind=operand_kind: '({}){}'.format(kind['type'], name)

ops_header.write('''
template<>
inline PairIdRefIdRef DecodeParam(const ConstIter &it, uint32_t &word)
{
  if(word >= it.size()) return {};
  
  PairIdRefIdRef ret = { Id::fromWord(it.word(word)), Id::fromWord(it.word(word+1)) };
  word += 2;
  return ret;
}

template<>
inline PairLiteralIntegerIdRef DecodeParam(const ConstIter &it, uint32_t &word)
{
  if(word >= it.size()) return {};
  
  PairLiteralIntegerIdRef ret = { it.word(word), Id::fromWord(it.word(word+1)) };
  word += 2;
  return ret;
}

template<>
inline PairIdRefLiteralInteger DecodeParam(const ConstIter &it, uint32_t &word)
{
  if(word >= it.size()) return {};
  
  PairIdRefLiteralInteger ret = { Id::fromWord(it.word(word)), it.word(word+1) };
  word += 2;
  return ret;
}
''')

tostrs = ''
tostr_decls = ''

# Second pass to declare operand parameter structs in ops helper header
for operand_kind in spirv['operand_kinds']:
    name = operand_kind['kind']

    if not operand_kind['has_params']:
        if operand_kind['category'] == 'ValueEnum':
            ops_header.write('inline uint16_t OptionalWordCount(const {0} val) {{ return val != {0}::Invalid ? 1 : 0; }}\n\n'.format(name))
        continue

    values = ''
    set_unset = ''
    word_count_cases = ''
    decode_cases = ''
    encode_cases = ''
    constructors = ''
    tostr_cases = ''

    value_enum = operand_kind['category'] == 'ValueEnum'
    bit_enum = operand_kind['category'] == 'BitEnum'

    used = []

    for value in operand_kind['enumerants']:
        params = ''
        assign = ''
        ret_assign = ''

        new_value = value['value'] not in used
        used.append(value['value'])

        if new_value and bit_enum:
            tostr_cases  += '  if(el.flags & {0}::{1})\n    ret += "{1}"'.format(name, value['enumerant'])

        if 'parameters' in value:
            # We want plain unions, so don't include strings
            if any([param['kind'] == 'LiteralString' for param in value['parameters']]):
                continue

            if new_value and value_enum:
                tostr_cases  += '    case {0}::{1}:\n      ret += '.format(name, value['enumerant'])

            member = ""
            param_name = operand_name(value['enumerant'])
            size = 0

            if new_value:
                if value_enum:
                    decode_cases += '    case {0}::{1}:\n'.format(name, value['enumerant'])
                    encode_cases += '    case {0}::{1}:\n'.format(name, value['enumerant'])
                else:
                    decode_cases += '  if(ret.flags & {0}::{1})\n  {{\n'.format(name, value['enumerant'])
                    encode_cases += '  if(param.flags & {0}::{1})\n  {{\n'.format(name, value['enumerant'])

            # if we only have one parameter, add its type to the set
            if len(value['parameters']) == 1:
                param = value['parameters'][0]
                size += kinds[param['kind']]['size']
                param_type = kinds[param['kind']]['type']
                member = "{} {};\n".format(param_type, param_name)

                if value_enum:
                    values += '  '
                    if new_value:
                        decode_cases += '  '
                        encode_cases += '  '

                values += "  " + member
                params += "{} {}Param".format(param_type, param_name)
                assign += " {0} = {0}Param;".format(param_name)
                ret_assign += "    ret.{0} = {0};\n".format(param_name)

                if new_value:
                    decode_cases += '    ret.{} = {};\n'.format(param_name, kinds[param['kind']]['from_words']('it.word(word)'))
                    encode_cases += '    {}\n'.format(kinds[param['kind']]['push_words']('param.{}'.format(param_name)))
                    if kinds[param['kind']]['is_id']:
                        tostr_cases += ' "(" + idName(el.{}) + ")"'.format(param_name)
                    else:
                        tostr_cases += ' "(" + ToStr(el.{}) + ")"'.format(param_name)

            # if we have multiple we need a separate struct for this thing
            else:
                struct_name = param_name[0].upper() + param_name[1:] + 'Params'
                member = "{} {};\n".format(struct_name, param_name)
                if value_enum:
                    values += '  '
                values += "  " + member

                struct_values = ''

                if new_value:
                    tostr_cases += ' "("'

                for i,param in enumerate(value['parameters']):
                    subparam_name = positional_names[i]
                    kind = kinds[param['kind']]
                    size += kind['size']
                    if 'name' in param:
                        subparam_name = operand_name(param['name'])
                    struct_values += "  {} {};\n".format(kind['type'], subparam_name)

                    if new_value:
                        if value_enum:
                            decode_cases += '  '
                            encode_cases += '  '
                        decode_cases += '    ret.{}.{} = {};\n'.format(param_name, subparam_name, kinds[param['kind']]['from_words']('it.word(word+{})'.format(i)))
                        encode_cases += '    {}\n'.format(kinds[param['kind']]['push_words']('param.{}.{}'.format(param_name, subparam_name)))
                        if kinds[param['kind']]['is_id']:
                            tostr_cases  += ' + idName(el.{}.{}) + '.format(param_name, subparam_name)
                        else:
                            tostr_cases  += ' + ToStr(el.{}.{}) + '.format(param_name, subparam_name)

                    assign += " {0}.{1} = {1};".format(param_name, subparam_name)
                    ret_assign += "    ret.{0}.{1} = {0}.{1};\n".format(param_name, subparam_name)
                    params += "{} {}".format(kind['type'], subparam_name)
                    if i != len(value['parameters'])-1:
                        params += ", "
                        tostr_cases += '", " '

                if new_value:
                    tostr_cases += '")"'

                header.write('''struct {struct_name}
{{
{struct_values}
}};

'''.format(struct_name = struct_name, struct_values = struct_values.rstrip()))

            if new_value:
                if value_enum:
                    decode_cases += '      word += {};\n'.format(size)
                    decode_cases += '      break;\n'
                    encode_cases += '      break;\n'
                    tostr_cases  += '; break;\n'
                else:
                    decode_cases += '    word += {};\n'.format(size)
                    decode_cases += '  }\n'
                    encode_cases += '  }\n'
                word_count_cases += '    case {}::{}: return {};\n'.format(name, value['enumerant'], size)

                constructors += '''template<>\nstruct {name}Param<{name}::{value}>
{{
  {member}
  {name}Param({params}) {{ {assign} }}
  operator {name}AndParamData()
  {{
    {name}AndParamData ret({name}::{value});
{ret_assign}
    return ret;
  }}
}};

'''.format(value=value['enumerant'], member=member.rstrip(), name=name, params=params, assign=assign, ret_assign=ret_assign.rstrip())

        if new_value and bit_enum:
            tostr_cases  += ' ", ";\n'


        set_unset += '''  void set{flag}({params}) {{ flags |= {name}::{flag};{assign} }}
  void unset{flag}() {{ flags &= ~{name}::{flag}; }}
'''.format(flag=value['enumerant'], name=name, params=params, assign=assign)

    if constructors != '':
        constructors = 'template<{name} val> struct {name}Param;\n\n'.format(name=name) + constructors

    # ValueEnums are set up as one or many pairs of enum/params, enum/params, etc. So we declare a struct for the pair
    # and declare an array if the op wants many
    if value_enum:
        tostrs += '''template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const rdcspv::{name}AndParamData &el)
{{
  rdcstr ret = ToStr(el.value);

  switch(el.value)
  {{
{tostr_cases}
    default:
      break;
  }}

  return ret;
}}

'''.format(name=name, tostr_cases=tostr_cases.rstrip())

        tostr_decls += '''template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const rdcspv::{name}AndParamData &el);'''.format(name=name)

        header.write('''struct {name}AndParamData
{{
  {name}AndParamData({name} v = {name}::Invalid) : value(v) {{}}
  {name} value;
  union
  {{
{values}
  }};
  
  operator {name}() const {{ return value; }}
  bool operator ==(const {name} v) const {{ return value == v; }}
}};

'''.format(name=name, values=values.rstrip()))

        ops_header.write('''{constructors}

template<>
inline {name}AndParamData DecodeParam(const ConstIter &it, uint32_t &word)
{{
  if(word >= it.size()) return {name}AndParamData();

  {name}AndParamData ret(({name})it.word(word));
  word++;
  switch(ret.value)
  {{
{decode_cases}
    default: break;
  }}
  return ret;
}}

inline void EncodeParam(rdcarray<uint32_t> &words, const {name}AndParamData &param)
{{
  words.push_back((uint32_t)param.value);
  switch(param.value)
  {{
{encode_cases}
    default: break;
  }}
}}

'''.format(name=name, value_name=operand_name(name), decode_cases=decode_cases.rstrip(),
           constructors=constructors, encode_cases=encode_cases.rstrip()))
        operand_kind['type'] = '{}AndParamData'.format(name)
    # BitEnums are set up with one bitmask, and then a series of parameters, so we declare a struct with an array
    elif bit_enum:
        tostrs += '''template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const rdcspv::{name}AndParamDatas &el)
{{
  rdcstr ret;
  
{tostr_cases}

  // remove trailing ", "
  if(ret.size() > 2)
    ret.erase(ret.size()-2, 2);

  return ret;
}}

'''.format(name=name, tostr_cases=tostr_cases.rstrip())

        tostr_decls += '''template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const rdcspv::{name}AndParamDatas &el);'''.format(name=name)

        header.write('''struct {name}AndParamDatas
{{
  {name}AndParamDatas({name} f = {name}::None) : flags(f) {{}}
  {name} flags;
{values}
  
  operator {name}() const {{ return flags; }}
  bool operator &(const {name} v) const {{ return bool(flags & v); }}
{set_unset}
}};

'''.format(name=name, values=values.rstrip(), set_unset=set_unset.rstrip()))

        ops_header.write('''template<>
inline {name}AndParamDatas DecodeParam(const ConstIter &it, uint32_t &word)
{{
  if(word >= it.size()) return {name}AndParamDatas();

  {name}AndParamDatas ret(({name})it.word(word));
  word++;
{decode_cases}
  return ret;
}}

inline void EncodeParam(rdcarray<uint32_t> &words, const {name}AndParamDatas &param)
{{
  words.push_back((uint32_t)param.flags);
{encode_cases}
}}

'''.format(name=name,  decode_cases=decode_cases.rstrip(), encode_cases=encode_cases.rstrip()))
        operand_kind['type'] = '{}AndParamDatas'.format(name)
    else:
        raise TypeError("unexpected operand kind {} with parameters".format(operand_kind['category']))

    ops_header.write('''inline uint16_t ExtraWordCount(const {name} {value_name})
{{
  switch({value_name})
  {{
{word_count_cases}
    default: break;
  }}
  return 0;
}}

'''.format(name = name, value_name = operand_name(name), word_count_cases = word_count_cases.rstrip()))

ops_header.write('''
inline uint16_t ExtraWordCount(const rdcstr &val)
{
  return uint16_t(val.size() / 4);
}

inline uint16_t OptionalWordCount(const rdcstr &val)
{
  if(val.empty()) return 0;
  return uint16_t(val.size() / 4) + 1;
}

inline uint16_t OptionalWordCount(const Id &val)
{
  return val != Id() ? 1 : 0;
}

inline uint16_t OptionalWordCount(const PairIdRefLiteralInteger &val)
{
  return val.first != Id() ? 2 : 0;
}

inline uint16_t OptionalWordCount(const PairLiteralIntegerIdRef &val)
{
  return val.second != Id() ? 2 : 0;
}

inline uint16_t OptionalWordCount(const PairIdRefIdRef &val)
{
  return val.first != Id() ? 2 : 0;
}

template<typename Type>
uint16_t MultiWordCount(const rdcarray<Type> &multiParams)
{
  uint16_t ret = 0;
  for(size_t i=0; i < multiParams.size(); i++)
    ret += sizeof(multiParams[i])/sizeof(uint32_t);
  return ret;
}
''')

###############################################################################
##
## Opcodes (declare enum / stringise)
##
###############################################################################

# Quickly preprocess, find parameters with duplicated names and disambiguate
for inst in spirv['instructions']:
    if 'operands' in inst:
        operands = inst['operands']

        duplicates = []

        for i,A in enumerate(operands):
            for j,B in enumerate(operands):
                if j <= i:
                    continue

                a = operand_name(A['name'] if 'name' in A else kinds[A['kind']]['def_name'])
                b = operand_name(B['name'] if 'name' in B else kinds[B['kind']]['def_name'])

                if a == b:
                    if i not in duplicates:
                        duplicates.append(i)
                    if j not in duplicates:
                        duplicates.append(j)

        if len(duplicates) > 0:
            for idx,arg in enumerate(duplicates):
                A = operands[arg]
                operands[arg]['name'] = operand_name(A['name'] if 'name' in A else kinds[A['kind']]['def_name']) + str(idx)

used = []
decl = ''
stringise = ''
op_structs = ''
op_decoder = ''
used_ids = ''
disassemble = ''

for inst in spirv['instructions']:
    decl += '  {} = {},\n'.format(inst['opname'][2:], inst['opcode'])

    if inst['opcode'] in used:
        continue

    stringise += '    STRINGISE_ENUM_CLASS({});\n'.format(inst['opname'][2:])

    result = -1
    resultCount = 0
    resultType = -1

    used_ids += '    case rdcspv::Op::{}:\n'.format(inst['opname'][2:])

    operands = []
    
    if 'operands' in inst:
        operands = inst['operands']
        last_operand = operands[-1]

        for i,operand in enumerate(operands):
            if operand['kind'] == 'IdResult':
                result = i+1
                resultCount += 1
            elif 'name' in operand and operand['name'][0:6] == 'result':
                resultCount += 1
            if operand['kind'] == 'IdResultType':
                resultType = i+1

    disassemble += '    case rdcspv::Op::{}:\n'.format(inst['opname'][2:])
    disassemble += '    {\n'

    if any([kinds[operand['kind']]['size'] is None for operand in operands]):
        op_struct = 'struct {}; // has operands with variable sizes\n\n'.format(inst['opname'])

        disassemble += '      OpDecoder decoded(it);\n'.format(inst['opname'][2:])

        if resultCount == 2:
            raise ValueError("Unexpected multiple results without decoded opcode")
        elif resultType > 0 and result > 0:
            disassemble += '      ret += declName(decoded.resultType, decoded.result) + " = ";\n'
        elif resultType > 0 and result == -1:
            raise ValueError("Unexpected result type without result")
        elif resultType == -1 and result > 0:
            disassemble += '      ret += idName(decoded.result) + " = ";\n'

        disassemble += '      ret += "{}(...)";\n'.format(inst['opname'][2:])
        disassemble += '      break;\n'
        disassemble += '    }\n'
    else:
        params = ''
        assign = ''
        member_decl = ''
        size_name = 'FixedWordSize'
        construct_size = 'FixedWordSize'
        size = 1 # opcode / wordcount packed
        all_size = 1 # size, but with all optionals included
        iter_init = '    memcpy(this, it.words(), sizeof(*this));'
        complex_type = False
        manual_init = '    this->op = OpCode;\n'
        manual_init += '    this->wordCount = (uint16_t)it.size();\n'
        oper_cast = '  operator Operation() const\n  {\n    rdcarray<uint32_t> words;\n'
        has_funcs = ''

        disassemble += '      Op{} decoded(it);\n'.format(inst['opname'][2:])
        
        if resultCount == 2:
            disassemble += '      ret += declName(decoded.resultType, decoded.result0) + ", " + idName(decoded.result1) + " = ";\n'
        elif resultType > 0 and result > 0:
            disassemble += '      ret += declName(decoded.resultType, decoded.result) + " = ";\n'
        elif resultType > 0 and result == -1:
            raise ValueError("Unexpected result type without result")
        elif resultType == -1 and result > 0:
            disassemble += '      ret += idName(decoded.result) + " = ";\n'

        disassemble += '      ret += "{}("'.format(inst['opname'][2:])

        disassemble_params = False

        if 'operands' in inst:
            for i,operand in enumerate(operands):
                kind = kinds[operand['kind']]

                if kind['has_params'] and not complex_type:
                    size_name = 'MinWordSize'
                    construct_size = 'MinWordSize'
                    complex_type = True
                    manual_init += '    uint32_t word = {};\n'.format(all_size)

                quantifier = ''

                if 'quantifier' in operand:
                    quantifier = operand['quantifier']
                    if not complex_type:
                        size_name = 'MinWordSize'
                        construct_size = 'MinWordSize'
                        complex_type = True
                        if quantifier == '*':
                            manual_init += '    uint32_t word = {};\n'.format(all_size)

                if kind['is_id']:
                    if quantifier == '*':
                        used_ids += '      for(size_t i=0; i < size-{0}; i++) callback(Id::fromWord(it.word({0}+i)), {1});\n'.format(all_size, 'true' if i+1==result else 'false')
                    else:
                        used_ids += '      callback(Id::fromWord(it.word({})), {});\n'.format(all_size, 'true' if i+1==result else 'false')

                if kind['size'] < 0:
                    size_name = 'MinWordSize'
                    construct_size = 'MinWordSize'
                    complex_type = True
                    manual_init += '    uint32_t word = {};\n'.format(all_size)

                opType,opName = (kind['type'], operand_name(operand['name'] if 'name' in operand else kind['def_name']))

                if i+1 != resultType and i+1 != result:
                    if quantifier == '*':
                        disassemble += ' + ParamsToStr(idName, decoded.{})'.format(opName)
                    else:
                        if opType == 'IdScope':
                            disassemble += ' + ToStr(Scope(constIntVal(decoded.{})))'.format(opName)
                        elif opType == 'IdMemorySemantics':
                            disassemble += ' + ToStr(MemorySemantics(constIntVal(decoded.{})))'.format(opName)
                        else:
                            disassemble += ' + ParamToStr(idName, decoded.{})'.format(opName)

                    if i+1 < len(operands):
                        disassemble += ' + ", "'

                    disassemble_params = True

                if quantifier == '?':
                    params += '{} {} = {}, '.format(opType, opName, kind['def_value'])
                elif quantifier == '*':
                    params += 'const rdcarray<{}> &{} = {{}}, '.format(opType, opName)
                else:
                    params += '{} {}, '.format(opType, opName)

                if quantifier == '*':
                    member_decl += '  rdcarray<{}> {};\n'.format(opType, opName)
                else:
                    member_decl += '  {} {};\n'.format(opType, opName)
                assign += '    this->{} = {};\n'.format(opName, opName)

                if operand['kind'] == 'LiteralString':
                    if quantifier == '*':
                        raise ValueError('operand {} in {} is string but has * quantifier'.format(opName, inst['opname']))
                    manual_init += '    this->{name} = DecodeParam<{type}>(it, word);\n'.format(name = opName, type = opType)
                    oper_cast += '    EncodeParam(words, {name});\n'.format(name = opName)
                    if quantifier == '?':
                        construct_size += ' + OptionalWordCount({})'.format(opName)
                        has_funcs += '  bool Has{name}() const {{ return wordCount > {idx}; }}\n'.format(idx = all_size, name = opName[0].upper() + opName[1:])
                    else:
                        construct_size += ' + ExtraWordCount({})'.format(opName)

                elif kind['has_params']:
                    if quantifier == '*':
                        raise ValueError('operand {} in {} has * quantifier and params'.format(opName, inst['opname']))
                    manual_init += '    this->{name} = DecodeParam<{type}>(it, word);\n'.format(name = opName, type = opType)
                    oper_cast += '    EncodeParam(words, {name});\n'.format(name = opName)
                    construct_size += ' + ExtraWordCount({})'.format(opName)
                elif quantifier == '*':
                    manual_init += '    this->{name} = MultiParam<{type}>(it, word);\n'.format(name = opName, type = opType)
                    construct_size += ' + MultiWordCount({})'.format(opName)
                    oper_cast += '    for(size_t i=0; i < {name}.size(); i++)\n'.format(name = opName)
                    oper_cast += '    {\n'
                    oper_cast += '      {push_words}\n'.format(push_words = kind['push_words']('{}[i]'.format(opName)))
                    oper_cast += '    }\n'
                elif quantifier == '?':
                    manual_init += '    this->{name} = (it.size() > {idx}) ? {value} : {def_value};\n'.format(name = opName, type = opType, idx = all_size, value = kind['from_words']('it.word({})'.format(all_size)), def_value=kind['def_value'])
                    construct_size += ' + OptionalWordCount({})'.format(opName)
                    oper_cast += '    if({name} != {def_value}) {push_words}\n'.format(name = opName, def_value=kind['def_value'], push_words = kind['push_words'](opName))
                    has_funcs += '  bool Has{name}() const {{ return wordCount > {idx}; }}\n'.format(idx = all_size, name = opName[0].upper() + opName[1:])
                else:
                    manual_init += '    this->{name} = {value};\n'.format(name = opName, type = opType, value = kind['from_words']('it.word({})'.format(all_size)))
                    oper_cast += '    {push_words}\n'.format(push_words = kind['push_words'](opName))

                if kind['size'] >= 0:
                    all_size += kind['size']
                else:
                    all_size += 1

                if quantifier == '':
                    size = all_size
        else:
            assign = '    // no operands'
            member_decl = '  // no operands'

        if complex_type:
            iter_init = manual_init.rstrip()
            oper_cast += '    return Operation(OpCode, words);\n  }\n'
        else:
            oper_cast = ''

        if params != '':
            params = params[0:-2]

        if disassemble_params:
            disassemble += ' + ")";\n'
        else:
            disassemble += ' ")";\n'
        disassemble += '      break;\n'
        disassemble += '    }\n'

        if has_funcs != '':
            has_funcs = '\n\n' + has_funcs

        op_struct = '''struct {name}
{{
  {name}(const ConstIter &it)
  {{
{iter_init}
  }}
  {name}({params})
      : op(Op::{opname})
      , wordCount({construct_size})
  {{
{assign}
  }}
{oper_cast}
  static constexpr Op OpCode = Op::{opname};
  static constexpr uint16_t {size_name} = {size}U;
  Op op;
  uint16_t wordCount;
{member_decl}{has_funcs}
}};

'''.format(opname=inst['opname'][2:], name=inst['opname'], params=params, iter_init=iter_init, assign=assign.rstrip(),
           member_decl=member_decl.rstrip(), size_name=size_name, construct_size=construct_size,
           oper_cast=oper_cast, size=size, has_funcs=has_funcs.rstrip())


    op_structs += op_struct

    # Sanity check that quantifiers only happen on final operands. Also if there are multiple they are all ?, not *
    if 'operands' in inst:
        operands = inst['operands']
        last_operand = operands[-1]

        for operand in operands:
            if operand != last_operand and 'quantifier' in operand and ('quantifier' not in last_operand or last_operand['quantifier'] != operand['quantifier'] or operand['quantifier'] != '?'):
                raise ValueError('quantifier on operand {} in {} but not on last operand'.format(operand['name'], inst['opname']))

    used_ids += '      break;\n'

    if result < 0:
        result = ' result = Id();'
    else:
        result = ' result = Id::fromWord(it.word({}));'.format(result)

    if resultType < 0:
        resultType = ' resultType = Id();'
    else:
        resultType = ' resultType = Id::fromWord(it.word({}));'.format(resultType)

    op_decoder += '    case rdcspv::Op::{}:{}{} break;\n'.format(inst['opname'][2:], result, resultType)

    used.append(inst['opcode'])

header.write('''enum class Op : uint16_t
{{
{decl}
  Max,
}};

'''.format(decl = decl))

ops_header.write('''
{op_structs}

template<typename T>
inline rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const T &el)
{{
  return ToStr(el);
}}

template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const Id &el);
template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const rdcstr &el);
template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const PairLiteralIntegerIdRef &el);
template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const PairIdRefLiteralInteger &el);
template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const PairIdRefIdRef &el);

{tostr_decls}

template<typename U>
inline rdcstr ParamsToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const rdcarray<U> &ids)
{{
  rdcstr ret = "{{";
  for(size_t i=0; i < ids.size(); i++)
  {{
    ret += ParamToStr(idName, ids[i]);
    if(i + 1 < ids.size())
      ret += ", ";
  }}
  ret += "}}";
  return ret;
}}

struct OpDecoder
{{
  OpDecoder(const ConstIter &it);

  static void ForEachID(const ConstIter &it, const std::function<void(Id,bool)> &callback);
  static rdcstr Disassemble(const ConstIter &it, const std::function<rdcstr(Id,Id)> &declName, const std::function<rdcstr(rdcspv::Id)> &idName, const std::function<uint32_t(Id)> &constIntVal);
  
  Op op;
  uint16_t wordCount;
  Id result;
  Id resultType;
}};
'''.format(op_structs = op_structs.rstrip(), tostr_decls = tostr_decls))

cpp.write('''template <>
rdcstr DoStringise(const rdcspv::Op &el)
{{
  BEGIN_ENUM_STRINGISE(rdcspv::Op);
  {{
{stringise}
  }}
  END_ENUM_STRINGISE();
}}

namespace rdcspv
{{

template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const Id &el)
{{
  return idName(el);
}}

template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const rdcstr &el)
{{
  return "\\"" + el + "\\"";
}}

template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const PairLiteralIntegerIdRef &el)
{{
  return StringFormat::Fmt("[%u, %s]", el.first, idName(el.second).c_str());
}}

template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const PairIdRefLiteralInteger &el)
{{
  return StringFormat::Fmt("[%s, %u]", idName(el.first).c_str(), el.second);
}}

template<>
rdcstr ParamToStr(const std::function<rdcstr(rdcspv::Id)> &idName, const PairIdRefIdRef &el)
{{
  return StringFormat::Fmt("[%s, %s]", idName(el.first).c_str(), idName(el.second).c_str());
}}

{tostrs}

void OpDecoder::ForEachID(const ConstIter &it, const std::function<void(Id,bool)> &callback)
{{
  size_t size = it.size();
  switch(it.opcode())
  {{
{used_ids}
    case Op::Max: break;
  }}
}}

rdcstr OpDecoder::Disassemble(const ConstIter &it, const std::function<rdcstr(Id,Id)> &declName, const std::function<rdcstr(rdcspv::Id)> &idName, const std::function<uint32_t(Id)> &constIntVal)
{{
  rdcstr ret;
  switch(it.opcode())
  {{
{disassemble}
    case Op::Max: break;
  }}
  return ret;
}}
  
OpDecoder::OpDecoder(const ConstIter &it)
{{
  op = it.opcode();
  wordCount = (uint16_t)it.size();
  switch(op)
  {{
{op_decoder}
    case Op::Max: break;
  }}
}}

}}; // namespace rdcspv

'''.format(stringise = stringise.rstrip(), op_decoder = op_decoder.rstrip(), used_ids = used_ids.rstrip(), disassemble = disassemble.rstrip(), tostrs = tostrs.rstrip()));

###############################################################################
##
## GLSL ext inst set (declare enum)
##
###############################################################################

decl = ''
stringise = ''

for inst in glsl450['instructions']:
    decl += '  {} = {},\n'.format(inst['opname'], inst['opcode'])
    stringise += '    STRINGISE_ENUM_CLASS({});\n'.format(inst['opname'])

header.write('''enum class GLSLstd450 : uint32_t
{{
{decl}
  Max,
  Invalid = ~0U,
}};

'''.format(decl = decl))

cpp.write('''template <>
rdcstr DoStringise(const rdcspv::GLSLstd450 &el)
{{
  BEGIN_ENUM_STRINGISE(rdcspv::GLSLstd450);
  {{
{stringise}
  }}
  END_ENUM_STRINGISE();
}}

template <>
rdcstr DoStringise(const rdcspv::Generator &el)
{{
  BEGIN_ENUM_STRINGISE(rdcspv::Generator);
  {{
{generator_tostr}
  }}
  END_ENUM_STRINGISE();
}}
'''.format(stringise = stringise.rstrip(), generator_tostr = generator_tostr.rstrip()))

header.write('''
}; // namespace rdcspv

DECLARE_STRINGISE_TYPE(rdcspv::GLSLstd450);
DECLARE_STRINGISE_TYPE(rdcspv::Generator);

''')

for operand_kind in spirv['operand_kinds']:
    if operand_kind['category'] == 'ValueEnum' or operand_kind['category'] == 'BitEnum':
        header.write('DECLARE_STRINGISE_TYPE(rdcspv::{});\n'.format(operand_kind['kind']))

ops_header.write('''
}; // namespace rdcspv
''')

header.close()
ops_header.close()
cpp.close()



from .util import *
from .capture import *
from .runner import *
from .analyse import *
from .testcase import *
from .shared.Texture_Zoo import *
from .shared.Mesh_Zoo import *
from .shared.Draw_Zoo import *
from .shared.Overlay_Test import *
from .shared.Buffer_Truncation import *
from .shared.Discard_Zoo import *



import rdtest
import renderdoc as rd


class VK_Int8_IBuffer(rdtest.TestCase):
    demos_test_name = 'VK_Int8_IBuffer'

    def check_capture(self):
        draw = self.find_draw("Draw")

        self.check(draw is not None)

        self.controller.SetFrameEvent(draw.eventId, False)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        postvs_data = self.get_postvs(draw, rd.MeshDataStage.VSOut, 0, draw.numIndices)

        ib = pipe.GetIBuffer()

        # Calculate the strip restart index for this index width
        striprestart_index = pipe.GetStripRestartIndex() & ((1 << (ib.byteStride*8)) - 1)

        # We don't check all of the output, we check a few key vertices to ensure they match up
        postvs_ref = {
            0: {
                'vtx': 0,
                'idx': 0,
                'gl_PerVertex_var.gl_Position': [-0.8, -0.2, 0.0, 1.0],
                'vertOut.col': [0.0, 1.0, 0.0, 1.0],
                'vertOut.uv': [0.0, 0.0, 0.0, 1.0],
            },
            4: {
                'vtx': 4,
                'idx': 4,
                'gl_PerVertex_var.gl_Position': [0.0, -0.2, 0.0, 1.0],
                'vertOut.col': [0.0, 1.0, 0.0, 1.0],
                'vertOut.uv': [0.0, 0.0, 0.0, 1.0],
            },
            8: {
                'idx': striprestart_index
            },
            9: {
                'vtx': 9,
                'idx': 8,
                'gl_PerVertex_var.gl_Position': [-0.8, 0.7, 0.0, 1.0],
                'vertOut.col': [0.0, 0.0, 1.0, 1.0],
                'vertOut.uv': [0.0, 0.0, 0.0, 1.0],
            },
        }

        self.check_mesh_data(postvs_ref, postvs_data)

        # Check that the rendered mesh is as expected
        out = pipe.GetOutputTargets()[0].resourceId
        for x in [x*0.01 for x in range(1, 100)]:
            self.check_pixel_value(out, x, 0.1, [0.2, 0.2, 0.2, 1.0])
            self.check_pixel_value(out, x, 0.5, [0.2, 0.2, 0.2, 1.0])

        self.check_pixel_value(out, 0.3, 0.25, [0.0, 1.0, 0.0, 1.0])
        self.check_pixel_value(out, 0.5, 0.25, [0.0, 1.0, 0.0, 1.0])
        self.check_pixel_value(out, 0.7, 0.25, [0.0, 1.0, 0.0, 1.0])
        self.check_pixel_value(out, 0.3, 0.75, [0.0, 0.0, 1.0, 1.0])
        self.check_pixel_value(out, 0.5, 0.75, [0.0, 0.0, 1.0, 1.0])
        self.check_pixel_value(out, 0.7, 0.75, [0.0, 0.0, 1.0, 1.0])



import renderdoc as rd
import rdtest
import os


class VK_RGP_Capture(rdtest.TestCase):
    demos_test_name = 'VK_Simple_Triangle'

    def check_capture(self):
        apiprops: rd.APIProperties = self.controller.GetAPIProperties()

        if not apiprops.rgpCapture:
            rdtest.log.print("RGP capture not tested")
            return

        path = self.controller.CreateRGPProfile(rd.CreateHeadlessWindowingData(100, 100))

        if os.path.exists(path) and os.path.getsize(path) > 100:
            rdtest.log.success("RGP capture created successfully")
        else:
            raise rdtest.TestFailureException("RGP capture failed")




import renderdoc as rd
import rdtest


class D3D12_Write_Subresource(rdtest.TestCase):
    demos_test_name = 'D3D12_Write_Subresource'

    def check_capture(self):
        draw = self.find_draw("Draw")

        self.controller.SetFrameEvent(draw.eventId, False)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        # Should be black around the sides, white in the centre
        self.check_pixel_value(pipe.GetOutputTargets()[0].resourceId, 0.05, 0.05, [0.0, 0.0, 0.0, 0.0])
        self.check_pixel_value(pipe.GetOutputTargets()[0].resourceId, 0.5, 0.5, [1.0, 1.0, 1.0, 1.0])

        rdtest.log.success("Picked values are as expected")


#!/usr/bin/env python3

import os
import io
import sys
import re
import xml.etree.ElementTree as ET

# on msys, use crlf output
nl = None
if sys.platform == 'msys':
    nl = "\r\n"

# Get the file, relative to this script's location (same directory)
# that way we're not sensitive to CWD
pathname = os.path.abspath(os.path.dirname(sys.argv[0])) + os.path.sep

# open the file for write
f = open(pathname + 'vk_dispatch_defs.h', mode='w', newline = nl)

# open XML registry
registry = ET.parse(pathname + 'vk.xml').getroot()

# f.write the file, starting with a template header
f.write('''
/******************************************************************************
 * The MIT License (MIT)
 *
 * Copyright (c) 2020-2021 Baldur Karlsson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 ******************************************************************************/

/******************************************************************************
 * Generated from Khronos's vk.xml:
'''.lstrip())

def prefix_star(line):
    if line == '':
        return ' *'
    else:
        return ' * ' + line

# Print the first two comments with the license
for comment in registry.findall('comment')[0:2]:
    f.write("\n".join([prefix_star(line.strip()) for line in comment.text.split('\n')]))

f.write('''
 ******************************************************************************/

// This file is autogenerated with gen_dispatch_table.py - any changes will be overwritten next time
// that script is run.
// $ ./gen_spirv_code.py

#pragma once

#include "official/vulkan.h"

// this file is autogenerated, so don't worry about clang-format issues
// clang-format off
'''.lstrip())

platform_defines = {}

# Cache the platform defines that protect each platform name
for plat in registry.findall('platforms/platform'):
    platform_defines[plat.attrib['name']] = plat.attrib['protect']

# Process all commands and categorise into instance or device
commands = {}
INSTANCE_CMD = 1
DEVICE_CMD = 2

# Some special cases we manually set
commands['vkCreateInstance'] = INSTANCE_CMD
commands['vkEnumerateInstanceVersion'] = INSTANCE_CMD
commands['vkEnumerateInstanceLayerProperties'] = INSTANCE_CMD
commands['vkEnumerateInstanceExtensionProperties'] = INSTANCE_CMD

import xml

for cmd in registry.findall('commands/command'):

    if 'alias' in cmd.attrib:
        name = cmd.attrib['name']
        alias = cmd.attrib['alias']
        if alias not in commands:
            raise ValueError('alias {} of {} defined, but {} is unknown'.format(name, alias, alias))

        commands[name] = commands[alias]
        continue

    name = cmd.find('proto/name').text
    if name in commands:
        continue

    first_param_type = cmd.find('param/type').text

    if first_param_type == 'VkInstance' or first_param_type == 'VkPhysicalDevice':
        commands[name] = INSTANCE_CMD
    elif first_param_type == 'VkDevice' or first_param_type == 'VkQueue' or first_param_type == 'VkCommandBuffer':
        commands[name] = DEVICE_CMD
    else:
        raise ValueError('type {} of first parameter to {} is unexpected'.format(first_param_type, name))

inst_commands = ""
dev_commands = ""
processed_commands = [] # some commands come from multiple extensions. Include them only in the first

def process_feature(root, name):
    global inst_commands, dev_commands, processed_commands

    inst = ""
    dev = ""

    for req in root.findall('require'):
        for cmd in req.findall('command'):
            function = cmd.attrib['name']

            if function in processed_commands:
                continue

            processed_commands.append(function)

            if function not in commands:
                raise ValueError('command {} referenced by {} is unknown'.format(function, name))

            table = commands[function]

            if table == INSTANCE_CMD:
                inst += '\n  PFN_{} {};'.format(function, function[2:])
            elif table == DEVICE_CMD:
                dev += '\n  PFN_{} {};'.format(function, function[2:])
            else:
                raise ValueError('command {} has unknown table type {}'.format(function, table))

    if 'platform' in root.attrib:
        if inst != "":
            inst = '\n#ifdef {plat}{inst}\n#endif // {plat}'.format(plat = platform_defines[root.attrib['platform']], inst = inst)
        if dev != "":
            dev = '\n#ifdef {plat}{dev}\n#endif // {plat}'.format(plat = platform_defines[root.attrib['platform']], dev = dev)

    if inst != "":
        inst_commands += "  // {name}{inst}\n\n".format(**locals())
    if dev != "":
        dev_commands += "  // {name}{dev}\n\n".format(**locals())

# Look at all features
for feat in registry.findall('feature'):
    # Only process vulkan features
    if 'api' in feat.attrib and feat.attrib['api'] == 'vulkan':
        process_feature(feat, feat.attrib['comment'])

# And all extensions (with KHR extensions sorted to the front)
def ext_sort(ext):
    if 'KHR' in ext.attrib['name']:
        return int(ext.attrib['number'])
    return 10000000 + int(ext.attrib['number'])


for ext in sorted(registry.findall('extensions/extension'), key=ext_sort):
    # Only process vulkan extensions
    if 'supported' in ext.attrib and ext.attrib['supported'] == 'vulkan':

        process_feature(ext, ext.attrib['name'])

inst_commands = inst_commands.strip()
dev_commands = dev_commands.strip()

f.write('''
struct VkInstDispatchTable
{{
  {inst_commands}
}};

struct VkDevDispatchTable
{{
  {dev_commands}

  // for consistency with macros, we declare the CreateDevice pointer here
  // even though it won't actually ever get used and is on the instance dispatch chain
  PFN_vkCreateDevice CreateDevice;
}};
'''.format(**locals()))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# RenderDoc documentation build configuration file, created by
# sphinx-quickstart on Thu May 12 16:59:09 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import re
import datetime

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

import struct

# path to module libraries for windows
if struct.calcsize("P") == 8:
    binpath = '../x64/'
else:
    binpath = '../Win32/'

# Prioritise release over development builds
sys.path.insert(0, os.path.abspath(binpath + 'Development/pymodules'))
sys.path.insert(0, os.path.abspath(binpath + 'Release/pymodules'))

# Add the build paths to PATH so renderdoc.dll can be located
os.environ["PATH"] += os.pathsep + os.path.abspath(binpath + 'Development/')
os.environ["PATH"] += os.pathsep + os.path.abspath(binpath + 'Release/')

# path to module libraries for linux
sys.path.insert(0, os.path.abspath('../build/lib'))

sys.path.insert(0, os.path.abspath('sphinx_exts'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc', 'sphinx_paramlinks']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'RenderDoc'
copyright = '{0}, Baldur Karlsson'.format(datetime.date.today().year)
author = 'Baldur Karlsson'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.

major_version = 123
minor_version = 999

with open('../renderdoc/api/replay/version.h') as f:
    for line in f:
        if line.find('#define RENDERDOC_VERSION_MAJOR') >= 0:
            major_version = line.split()[2]
        if line.find('#define RENDERDOC_VERSION_MINOR') >= 0:
            minor_version = line.split()[2]

version = '{0}.{1}'.format(major_version, minor_version)
# The full version, including alpha/beta/rc tags.
release = 'v{0}'.format(version)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'sphinx_exts', 'include']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
html_title = 'RenderDoc documentation'

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
#html_last_updated_fmt = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = False

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr', 'zh'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# 'ja' uses this config value.
# 'zh' user can custom change `jieba` dictionary path.
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'renderdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'RenderDoc.tex', 'RenderDoc Documentation',
     'Baldur Karlsson', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'renderdoc', 'RenderDoc Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'RenderDoc', 'RenderDoc Documentation',
     author, 'RenderDoc', 'One line description of project.',
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False

# custom theme based on sphinx_rtd_theme
if os.path.isdir('sphinx_rtd_theme_chm_friendly'):
    html_theme = "sphinx_rtd_theme_chm_friendly"
    html_theme_path = ["."]
else:
    html_theme = "sphinx_rtd_theme"

html_context = {
    'show_source': False,
    'html_show_sourcelink': False,
    'display_github': True,
    'github_user': 'baldurk',
    'github_repo': 'renderdoc',
    'github_version': 'v{0}'.format(version),
    'conf_py_path': '/docs/',
}

# We need 1.5 and above for the htmlhelp links to be handled properly without
# needing separate ugly _blank links. If you don't care about that, you can
# disable this
if(tags.has('htmlhelp')):
    print("**** We require sphinx 1.5 for htmlhelp build to have the fix for issue #2550 ****")
    needs_sphinx = '1.5'

def maybe_skip_member(app, what, name, obj, skip, options):
    # Hide these SWIG internals
    if name == "this" or name == "thisown":
        return True
    # Allow hiding free module functions, or only showing free module functions
    if 'exclude-members' in options and what == "module":
        if 'free_functions__' in options['exclude-members'] and 'built-in function' in repr(obj):
            return True
        if 'non_free_functions__' in options['exclude-members'] and 'built-in function' not in repr(obj):
            return True
    # Allow hiding enum constant members (i.e. int constants). These can then be documented explicitly
    # as we don't have a way in SWIG to attach docstrings to constants directly.
    if 'exclude-members' in options and 'enum_constants__' in options['exclude-members'] and isinstance(obj, int):
        return True
    if 'exclude-members' in options and 'properties__' in options['exclude-members'] and 'getset_desc' in str(type(obj)):
        return True
    # Allow arbitrary globbing as a hack to exclude or include members
    if 'exclude-members' in options:
        for exclude in options['exclude-members']:
            # Look for a hack that describes a name match
            if exclude.startswith('name_match__'):
                match = exclude.replace('name_match__', '')

                include_only = False

                # see if it wants to include only matches, or exclude matches (default)
                if match.startswith('include_only__'):
                    match = match.replace('include_only__', '')
                    include_only = True

                objname = ""
                if '__qualname__' in dir(obj):
                    objname = obj.__qualname__
                else:
                    try:
                        objname = obj.__name__
                    except AttributeError:
                        objname = obj.__class__.__name__
                ismatch = False

                # see if we're matching a prefix, or doing just a glob
                if match.startswith('startswith__'):
                    match = match.replace('startswith__', '')
                    ismatch = objname.startswith(match)

                if match.startswith('in__'):
                    match = match.replace('in__', '')
                    ismatch = match in objname

                # if we want to include only matches and it didn't match, skip this
                if include_only and not ismatch:
                    return True

                # If we want to exclude matches and it DID match, skip
                if not include_only and ismatch:
                    return True
    return None

def build_finished(app, exception):
    import renderdoc as rd
    import qrenderdoc as qrd

    from sphinx.domains.python import PythonDomain
    from sphinx.errors import SphinxError

    if exception is not None:
        return

    print(rd)
    print(qrd)

    # Get list of documented/indexed python objects
    pydomain = app.env.get_domain('py')
    if not hasattr(pydomain, 'objects'):
        print("WARNING: Sphinx version is too old to check objects validity. Upgrade to at least Sphinx 2.1.0")
        return
    objs = pydomain.objects

    # Enumerate the namespaced objects in both modules
    items = []
    for module_name in ['renderdoc', 'qrenderdoc']:
        module = sys.modules[module_name]
        entries = dir(module)
        for item in dir(module):
            if 'INTERNAL:' not in str(module.__dict__[item].__doc__):
                items.append('{}.{}'.format(module_name, item))

    items = set(filter(lambda i: re.search('__|SWIG|ResourceId_Null|rdcfixedarray_of|rdcarray_of|Structured.*List', i) is None, items))

    # Remove any documented/indexed python objects
    items -= set(objs.keys())

    # Print an error if any remain
    if len(items) > 0:
        items = sorted(list(items))
        raise SphinxError("These {} global classes/functions are not included in the documentation index:\n* {}".format(len(items), '\n* '.join(items)))

    print("All python objects are linked in the documentation.")

def setup(app):
    app.connect('autodoc-skip-member', maybe_skip_member)
    app.connect('build-finished', build_finished)



import struct
import math
import renderdoc as rd
import rdtest


class D3D12_Parameter_Zoo(rdtest.TestCase):
    demos_test_name = 'D3D12_Parameter_Zoo'

    def check_capture(self):
        draw = self.find_draw("Color Draw")

        self.check(draw is not None)

        draw = draw.next

        self.controller.SetFrameEvent(draw.eventId, False)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        self.check_pixel_value(pipe.GetOutputTargets()[0].resourceId, 0.5, 0.5, [0.0, 1.0, 0.0, 1.0])

        rdtest.log.success("Captured loaded with color as expected")

        postvs_data = self.get_postvs(draw, rd.MeshDataStage.VSOut, 0, draw.numIndices)

        postvs_ref = {
            0: {
                'vtx': 0,
                'idx': 0,
                'SV_POSITION': [-0.5, -0.5, 0.0, 1.0],
                'COLOR': [0.0, 1.0, 0.0, 1.0],
                'TEXCOORD': [0.0, 0.0],
            },
            1: {
                'vtx': 1,
                'idx': 1,
                'SV_POSITION': [0.0, 0.5, 0.0, 1.0],
                'COLOR': [0.0, 1.0, 0.0, 1.0],
                'TEXCOORD': [0.0, 1.0],
            },
            2: {
                'vtx': 2,
                'idx': 2,
                'SV_POSITION': [0.5, -0.5, 0.0, 1.0],
                'COLOR': [0.0, 1.0, 0.0, 1.0],
                'TEXCOORD': [1.0, 0.0],
            },
        }

        self.check_mesh_data(postvs_ref, postvs_data)

        rdtest.log.success("Mesh data is correct")

        tex = rd.TextureDisplay()
        tex.overlay = rd.DebugOverlay.Drawcall
        tex.resourceId = pipe.GetOutputTargets()[0].resourceId

        out: rd.ReplayOutput = self.controller.CreateOutput(rd.CreateHeadlessWindowingData(100, 100),
                                                            rd.ReplayOutputType.Texture)

        out.SetTextureDisplay(tex)

        out.Display()

        overlay_id = out.GetDebugOverlayTexID()

        v = pipe.GetViewport(0)

        self.check_pixel_value(overlay_id, int(0.5 * v.width), int(0.5 * v.height), [0.8, 0.1, 0.8, 1.0],
                               eps=1.0 / 256.0)

        out.Shutdown()

        rdtest.log.success("Overlay color is as expected")


import renderdoc as rd
import threading
import time

# This sample is intended as an example of how to do remote capture and replay
# as well as using device protocols to automatically enumerate remote targets.
#
# It is not complete since it requires filling in with custom logic to select
# the executable and trigger the capture at the desired time
raise RuntimeError("This sample should not be run directly, read the source")

rd.InitialiseReplay(rd.GlobalEnvironment(), [])

protocols = rd.GetSupportedDeviceProtocols()

print(f"Supported device protocols: {protocols}")

# Protocols are optional - they allow automatic detection and management of
# devices.
if protocol_to_use is not None:
    # the protocol must be supported
    if protocol_to_use not in protocols:
        raise RuntimeError(f"{protocol_to_use} protocol not supported")

    protocol = rd.GetDeviceProtocolController(protocol_to_use)

    devices = protocol.GetDevices()

    if len(devices) == 0:
        raise RuntimeError(f"no {protocol_to_use} devices connected")

    # Choose the first device
    dev = devices[0]
    name = protocol.GetFriendlyName(dev)

    print(f"Running test on {dev} - named {name}")

    URL = protocol.GetProtocolName() + "://" + dev

    # Protocols can enumerate devices which are not supported. Capture/replay
    # is not guaranteed to work on these devices
    if not protocol.IsSupported(URL):
        raise RuntimeError(f"{dev} doesn't support capture/replay - too old?")

    # Protocol devices may be single-use and not support multiple captured programs
    # If so, trying to execute a program for capture is an error
    if not protocol.SupportsMultiplePrograms(URL):
        # check to see if anything is running. Just use the URL
        ident = rd.EnumerateRemoteTargets(URL, 0)

        if ident != 0:
            raise RuntimeError(f"{name} already has a program running on {ident}")
else:
    # If you're not using a protocol then the URL can simply be a hostname.
    # The remote server must be running already - how that is done is up
    # to you. Everything else will work the same over a normal TCP connection
    protocol = None
    URL = hostname

# Let's try to connect
status,remote = rd.CreateRemoteServerConnection(URL)

if status == rd.ReplayStatus.NetworkIOFailed and protocol is not None:
    # If there's just no I/O, most likely the server is not running. If we have
    # a protocol, we can try to start the remote server
    print("Couldn't connect to remote server, trying to start it")

    status = protocol.StartRemoteServer(URL)

    if status != rd.ReplayStatus.Succeeded:
        raise RuntimeError(f"Couldn't launch remote server, got error {str(status)}")

    # Try to connect again!
    status,remote = rd.CreateRemoteServerConnection(URL)

if status != rd.ReplayStatus.Succeeded:
    raise RuntimeError(f"Couldn't connect to remote server, got error {str(status)}")

# We now have a remote connection. This works regardless of whether it's a device
# with a protocol or not. In fact we are done with the protocol at this point
protocol = None

print("Got connection to remote server")

# GetHomeFolder() gives you a good default path to start with.
# ListFolder() lists the contents of a folder and can recursively
# browse the remote filesystem.
home = remote.GetHomeFolder()
paths = remote.ListFolder(home)

print(f"Executables in home folder '{home}':")

for p in paths:
    print("  - " + p.filename)

# Select your executable, perhaps hardcoded or browsing using the above
# functions
exe,workingDir,cmdLine,env,opts = select_executable()

print(f"Running {exe}")

result = remote.ExecuteAndInject(exe, workingDir, cmdLine, env, opts)

if result.status != rd.ReplayStatus.Succeeded:
    remote.ShutdownServerAndConnection()
    raise RuntimeError(f"Couldn't launch {exe}, got error {str(result.status)}")

# Spin up a thread to keep the remote server connection alive while we make a capture,
# as it will time out after 5 seconds of inactivity
def ping_remote(remote, kill):
    success = True
    while success and not kill.is_set():
        success = remote.Ping()
        time.sleep(1)

kill = threading.Event()
ping_thread = threading.Thread(target=ping_remote, args=(remote,kill))
ping_thread.start()

# Create target control connection
target = rd.CreateTargetControl(URL, result.ident, 'remote_capture.py', True)

if target is None:
    kill.set()
    ping_thread.join()
    remote.ShutdownServerAndConnection()
    raise RuntimeError(f"Couldn't connect to target control for {exe}")

print("Connected - waiting for desired capture")

# Wait for the capture condition we want
capture_condition()

print("Triggering capture")

target.TriggerCapture(1)

# Pump messages, keep waiting until we get a capture message. Time out after 30 seconds
msg = None
start = time.clock()
while msg is None or msg.type != rd.TargetControlMessageType.NewCapture:
    msg = target.ReceiveMessage(None)

    if time.clock() - start > 30:
        break

# Close the target connection, we're done either way
target.Shutdown()
target = None

# Stop the background ping thread
kill.set()
ping_thread.join()

# If we didn't get a capture, error now
if msg.type != rd.TargetControlMessageType.NewCapture:
    remote.ShutdownServerAndConnection()
    raise RuntimeError("Didn't get new capture notification after triggering capture")

cap_path = msg.newCapture.path
cap_id = msg.newCapture.captureId

print(f"Got new capture at {cap_path} which is frame {msg.newCapture.frameNumber} with {msg.newCapture.api}")

# We could save the capture locally
# remote.CopyCaptureFromRemote(cap_path, local_path, None)


# Open a replay. It's recommended to set no proxy preference, but you could
# call remote.LocalProxies and choose an index.
#
# The path must be remote - if the capture isn't freshly created then you need
# to copy it with remote.CopyCaptureToRemote()
status,controller = remote.OpenCapture(rd.RemoteServer.NoPreference, cap_path, rd.ReplayOptions(), None)

if status != rd.ReplayStatus.Succeeded:
    remote.ShutdownServerAndConnection()
    raise RuntimeError(f"Couldn't open {cap_path}, got error {str(result.status)}")

# We can now use replay as normal.
#
# The replay is tunnelled over the remote connection, so you don't have to keep
# pinging the remote connection while using the controller. Use of the remote
# connection and controller can be interleaved though you should only access
# them from one thread at once. If they are both unused for 5 seconds though,
# the timeout will happen, so if the controller is idle it's advisable to ping
# the remote connection

sampleCode(controller)

print("Shutting down")

controller.Shutdown()

# We can still use remote here - e.g. capture again, replay something else,
# save the capture, etc

remote.ShutdownServerAndConnection()

rd.ShutdownReplay()



import struct
from typing import List
import renderdoc as rd
import rdtest


class GL_Per_Type_Tex_Units(rdtest.TestCase):
    demos_test_name = 'GL_Per_Type_Tex_Units'

    def check_capture(self):
        draw = self.find_draw("Draw")

        self.controller.SetFrameEvent(draw.eventId, False)

        pipe: rd.PipeState = self.controller.GetPipelineState()

        bind: rd.ShaderBindpointMapping = pipe.GetBindpointMapping(rd.ShaderStage.Fragment)
        texs: List[rd.BoundResourceArray] = pipe.GetReadOnlyResources(rd.ShaderStage.Fragment)

        if len(bind.readOnlyResources) != 2:
            raise rdtest.TestFailureException(
                "Expected 2 textures bound, not {}".format(len(bind.readOnlyResources)))

        if bind.readOnlyResources[0].bind != 2:
            raise rdtest.TestFailureException(
                "First texture should be on slot 2, not {}".format(bind.readOnlyResources[0].bind))

        id = texs[2].resources[0].resourceId

        tex_details = self.get_texture(id)
        res_details = self.get_resource(id)

        if res_details.name != "Red 2D":
            raise rdtest.TestFailureException("First texture should be Red 2D texture, not {}".format(res_details.name))

        if tex_details.dimension != 2:
            raise rdtest.TestFailureException(
                "First texture should be 2D texture, not {}".format(tex_details.dimension))

        if tex_details.width != 8 or tex_details.height != 8:
            raise rdtest.TestFailureException(
                "First texture should be 8x8, not {}x{}".format(tex_details.width, tex_details.height))

        data = self.controller.GetTextureData(id, rd.Subresource(0, 0, 0))
        first_pixel = struct.unpack_from("BBBB", data, 0)

        if not rdtest.value_compare(first_pixel, (255, 0, 0, 255)):
            raise rdtest.TestFailureException("Texture should contain red, not {}".format(first_pixel))

        rdtest.log.success("First texture is as expected")

        if bind.readOnlyResources[1].bind != 3:
            raise rdtest.TestFailureException(
                "First texture should be on slot 3, not {}".format(texs[0].bindPoint.bind))

        id = texs[3].resources[0].resourceId

        tex_details = self.get_texture(id)
        res_details = self.get_resource(id)

        if res_details.name != "Green 3D":
            raise rdtest.TestFailureException(
                "First texture should be Green 3D texture, not {}".format(res_details.name))

        if tex_details.dimension != 3:
            raise rdtest.TestFailureException(
                "First texture should be 3D texture, not {}".format(tex_details.dimension))

        if tex_details.width != 4 or tex_details.height != 4 or tex_details.depth != 4:
            raise rdtest.TestFailureException(
                "First texture should be 4x4x4, not {}x{}x{}".format(tex_details.width, tex_details.height,
                                                                     tex_details.depth))

        data = self.controller.GetTextureData(id, rd.Subresource(0, 0, 0))
        first_pixel = struct.unpack_from("BBBB", data, 0)

        if not rdtest.value_compare(first_pixel, (0, 255, 0, 255)):
            raise rdtest.TestFailureException("Texture should contain green, not {}".format(first_pixel))

        rdtest.log.success("Second texture is as expected")

        self.check_pixel_value(pipe.GetOutputTargets()[0].resourceId, 0.5, 0.5, [1.0, 1.0, 0.0, 0.2])

        rdtest.log.success("Picked value is as expected")
